'use strict'

const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)
const buildRichMessage = require('./rich-messages')
const request = require('client-request')
const asynclib = require('async')

module.exports = init

// This should be more than enough time
const REQUEST_TIMEOUT = 30000

function init (context) {
  return new MessageAction(context)
}

class MessageAction {
  constructor (context) {
    this._context = context
    this.type = null
  }

  run (params, cb) {
    // find action configurations
    const actionConfigs = this.getActionConfigs(params)
    if (!actionConfigs || actionConfigs.length === 0) return setImmediate(cb)

    this.type = params.type

    // Don't force these to run in series because one might be slow, but don't
    // allow unlimited parallelism
    asynclib.mapLimit(actionConfigs, 3, (actionConfig, done) => {
      if (!this.shouldRun(params, actionConfig)) return done()

      Logger.info(`Sending ${this.type} message for ${params.view.name || params.event} to ${actionConfig.url}`)

      const urlParts = actionConfig.url.split('?')
      var message

      try {
        message = buildRichMessage({
          format: this.type,
          consoleURL: params.consoleURL,
          data: params.data,
          type: params.event,
          view: params.view,
          agents: params.agents,
          artifacts: params.artifacts
        })
      } catch (err) {
        // Don't pass error along to asynclib because it will bail on other
        // message generation which we don't want
        Logger.error(`Error building ${this.type} message for ${params.event}: ${err}`)
        return done()
      }

      // If the action declined to return a message, it was filtered for some reason
      if (message == null) {
        return
      }

      const body = this.type === 'teams'
        ? JSON.stringify(message)
        : (typeof message === 'string') ? {text: message} : message
      const reqParams = {
        uri: urlParts[0],
        method: 'POST',
        body: body,
        timeout: REQUEST_TIMEOUT
      }

      if (this.type === 'teams') {
        reqParams.headers = {
          'Content-Type': 'application/json',
          'Content-Length': body.length
        }
      }

      request(reqParams, (err, res) => {
        if (err) {
          Logger.warn(`Error posting to ${this.type} at ${urlParts[0]}: ${err.message}`)
        }
        Logger.debug(`${this.type} results`, err, res.statusCode)
        done(null, res.statusCode)
      })
    }, (err, codes) => {
      cb(err, codes)
    })
  }

  // This only needs to be done for vulnerabilities at the moment.
  shouldRun (params, actionConfig) {
    switch (params.event) {
      case 'nsolid-agent-exit':
        const exitType = params.view.exitType || 'all'
        const hasStack = !!(params.data && params.data.error && params.data.error.stack)

        // Only agents 2.4 and newer supply exitCode, just report on them regardless prior to that
        if (params.data.hasOwnProperty('exitCode')) {
          if (params.data.exitCode === 0) {
            if (exitType === 'abnormal') return false
          } else {
            if (exitType === 'normal') return false
          }
        }

        if (params.view.onlyExceptions && !hasStack) {
          return false
        }

        return true
      default:
        return true
    }
  }

  getActionConfigs (params) {
    return params.view.actions.filter(action => {
      return action && action.type === params.type && !!action.url
    })
  }
}
