'use strict'

const fs = require('fs')
const path = require('path')
const RevocableEventEmitter = require('./RevocableEventEmitter')
const async = require('async')
const util = require('./util')
const snapshotSummaries = require('./snapshot-summaries')
const Logger = util.getLogger(__dirname, __filename)
const zlib = require('zlib')

const writeFileOpts = {
  defaultEncoding: null,
  flags: 'a'
}

// This is always on in production but can be disabled
const COMPRESS_ASSETS = true

exports.create = create
exports.mkdirp = fs.mkdir

function create (context, params, onComplete) {
  try {
    return new Asset(context, params, onComplete)
  } catch (err) {
    Logger.error('Could not create asset.', err, params)
    return null
  }
}

class Asset extends RevocableEventEmitter {
  constructor (context, params, onComplete) {
    super()
    this.bulkId = params.bulkId
    this.agentId = params.agentId
    this.viewId = params.metadata ? params.metadata.viewId : undefined
    this.viewName = params.metadata && params.metadata.viewName ? params.metadata.viewName : undefined
    this.reason = params.metadata && params.metadata.reason ? params.metadata.reason : 'Unknown'
    this.type = params.type || params.command
    this.size = 0
    this._onComplete = onComplete || function () {}
    this._context = context
    this._paused = true
    this._stream = null
    this._writeQueue = []
    this._complete = false
    this._compressed = COMPRESS_ASSETS

    // 2016-07-22T17:56:02.077Z -> 2016-07-22_17-56-02-077Z
    const basename = new Date().toISOString()
      .replace('T', '_')
      .replace('.', '-')
      .replace(/:/g, '-')

    let ext = 'dat'
    if (this.type === 'packages') ext = 'packages.json'
    else if (this.type === 'profile') ext = 'cpuprofile'
    else if (this.type === 'snapshot') ext = 'heapsnapshot'
    else if (this.type === 'snapshot-summary') ext = this.type

    if (this.type === 'profile' && params.duration) {
      this.duration = params.duration
    }

    this.assetId = path.join(
      this.type,
      this.agentId,
      `${basename}.${ext}`
    )

    this._filepath = path.join(
      context.config.data.dir,
      'assets',
      this.assetId
    )

    if (!this._context || !this.bulkId || !this.type || !this.agentId) {
      throw new Error('Context, bulkId, agentId, and type required for Asset.')
    }

    Logger.debug('creating Asset', params, `${this}`)

    exports.mkdirp(path.dirname(this._filepath), { recursive: true }, err => {
      if (err) {
        return this.cancel(err)
      }

      if (this._compressed) {
        const compressedPath = this._filepath + '.gz'
        const fileStream = fs.createWriteStream(compressedPath, writeFileOpts)
        const gzipStream = zlib.createGzip()

        // propagate the file open event
        fileStream.on('open', () => { gzipStream.emit('open') })

        gzipStream.pipe(fileStream)
        this._stream = gzipStream
      } else {
        this._stream = fs.createWriteStream(this._filepath, writeFileOpts)
      }
      this.emit('stream')

      this._stream.on('open', () => {
        this.flush()
      })

      this._stream.on('error', err => {
        Logger.error('Error writing file to disk', err)
        this.cancel(err)
      })

      this._stream.on('drain', () => {
        this.flush()
      })

      this._stream.on('finish', () => {
        this.emit('end')
      })
    })
  }

  getBulkId () {
    return this.bulkId
  }

  getFileName () {
    return this._filepath
  }

  getDownloadFileName () {
    return `${this.agentId}-${path.basename(this.getAssetId())}`
  }

  getAssetId () {
    return this.assetId
  }

  toString () {
    return `Asset{filePath:${this._filepath}, complete:${this._complete}}`
  }

  flush () {
    let continueWrites = true

    async.whilst(() => {
      return continueWrites && !!this._writeQueue.length
    }, next => {
      // If the previous we haven't hit HWM and still have queued data,
      // write the next chunk in order. `next` must be called asynchronously
      // so `continueWrites` new value is available after assignment.
      const data = this._writeQueue.shift()
      continueWrites = this._stream.write(data, () => setImmediate(next))
    }, (err, res) => {
      if (err) {
        return this.cancel(err)
      }

      if (continueWrites && this._writeQueue.length) {
        // We may get additional queued data between iterator and callback.
        // Keep on flushing, keep on flushing...
        this.flush()
        return
      }

      // We've already written all the data.
      if (!this._writeQueue.length && this._complete) {
        this.cancel()
        return
      }

      // Flush has finished, we've caught up and can handle writing new data
      // immediately again.
      if (continueWrites) {
        this._paused = false
      }
    })
  }

  addData (envelope, data) {
    if (!this._context) {
      return
    }

    if (data) {
      this.size += data.length

      // We either don't have a stream yet, or we've hit HWM and haven't
      // drained.  If unflushed segments remain in the queue, continue to queue
      // so they are written in order.
      if (!this._stream || this._paused || this._writeQueue.length > 0) {
        this._writeQueue.push(data)
      } else if (!this._stream.write(data)) {
        this._paused = true
      }
    }

    // When we receive complete: true, it indicates no further data will be
    // sent; the file is ended.
    if (envelope.complete) {
      this._complete = true

      // We've already written all the data.
      if (this._stream && !this._writeQueue.length) {
        this.cancel()
      }
    }
  }

  // cancel also used as the normal "close" path, err = null in that case
  cancel (err) {
    if (this._stream) {
      this._stream.once('close', onClosed)
      if (this._compressed) {
        this._stream.flush()
      }
      this._stream.end()
    }
    this._stream = null

    const self = this

    function onClosed () {
      if (self._onComplete) {
        self._onComplete(err)
        Logger.debug(`asset complete for ${self.getBulkId()} ${self.getAssetId()}`, err)
      }
      self._onComplete = null

      if (err) {
        self.emit('complete', err)
        return
      }

      if (self.type === 'snapshot') {
        self.addSummary(err => {
          self.emit('complete', err)
        })
      } else {
        self.emit('complete')
      }
    }
  }

  addSummary (cb) {
    snapshotSummaries.addSummaryAsset(this._context, this, (err, summaryAsset) => {
      if (err) return cb(err)
      this.summaryAsset = summaryAsset.bulkId
      cb(null)
    })
  }
}
