'use strict'

// test this list against current nsolid values by running:
// nsolid lib/fields-test-live.js

exports.create = create
exports.flattenObject = flattenObject
exports.splitPackageVersion = splitPackageVersion

const FIELD_RANGE_RESET_SECONDS = 60

const d3Scale = require('d3-scale')
const semver = require('semver')

const util = require('./util')

const Logger = util.getLogger(__dirname, __filename)

// create a new field manager
function create (context) {
  const fieldDomains = util.readSchema('fields')
  if (fieldDomains == null) return null

  const fieldManager = new FieldManager(context, fieldDomains)

  setInterval(() => fieldManager.resetRanges(), FIELD_RANGE_RESET_SECONDS * 1000).unref()

  return fieldManager
}

// container for all the known fields
class FieldManager {
  constructor (context, domains) {
    this._context = context
    this._fields = {}
    this._domains = Object.keys(domains)

    for (let domain in domains) {
      for (let name in domains[domain]) {
        const fieldDescriptor = {
          domain: domain,
          name: name,
          type: domains[domain][name].type,
          desc: domains[domain][name].desc,
          range: true,
          editor: domains[domain][name].editor,
          shortName: domains[domain][name].shortName,
          displayName: domains[domain][name].displayName,
          unit: domains[domain][name].unit
        }

        // range value of a field defaults to true, but can be explicitly
        // set to false; false means don't bother tracking the range, eg
        // for agent id
        if (domains[domain][name].range === false) {
          fieldDescriptor.range = false
        }

        // newly initialized fields will have their ranges set to appropriate
        // initial values unless fieldDescriptor.range == false, in which
        // case range will be set to null
        let field
        switch (fieldDescriptor.type) {
          case 'number': field = new FieldNumber(fieldDescriptor); break
          case 'string': field = new FieldString(fieldDescriptor); break
          case 'boolean': field = new FieldBoolean(fieldDescriptor); break
          case 'package': field = new FieldPackage(fieldDescriptor); break
        }

        if (field == null) {
          Logger.warn(`invalid field type for ${fieldDescriptor.domain} ${fieldDescriptor.name}`)
          continue
        }

        this._fields[fieldDescriptor.name] = field
      }
    }
  }

  // return all the fields, in an object with the field name as property values
  get fields () {
    return JSON.parse(JSON.stringify(this._fields))
  }

  // return fields in a , in an object with field name as property values
  fieldsInDomain (domain) {
    const result = {}

    for (let name in this._fields) {
      const field = this._fields[name]
      if (field.domain === domain) result[name] = field
    }

    return result
  }

  // return a single field
  getField (name) {
    return this._fields[name]
  }

  resetRanges () {
    Logger.debug('resetting field ranges')
    // in case agentManager isn't hooked up in a test
    if (this._context.agentManager == null) {
      Logger.warn('no agentManager, so not resetting ranges')
      return
    }

    // capture existing ranges and reset ranges to initial "empty" values
    const originalRanges = new Map()
    for (let name in this._fields) {
      const field = this._fields[name]
      if (field.range == null) continue

      originalRanges.set(field, field.range)

      field.prepResetRange()
      field.range = field.initialRange
    }

    // update ranges for all existing agents
    for (let agent of this._context.agentManager.list()) {
      if (agent.info == null || agent.metrics == null) continue
      this.updateRanges('info', agent.info, true)
      this.updateRanges('metrics', agent.metrics, true)

      if (agent.vulns == null) continue
      const vulnIds = agent.vulns.map(vuln => vuln.id)
      this.updateRanges('packages', { vuln: vulnIds }, true)
    }

    // send events for changes
    for (let field of originalRanges.keys()) {
      const oldRange = originalRanges.get(field)
      const newRange = field.range
      if (JSON.stringify(oldRange) === JSON.stringify(newRange)) continue

      this.sendRangeChangedEvent(field)
    }
  }

  // update the live ranges for fields given values in the object
  updateRanges (domain, object, quiet) {
    // only info objects need to be flattened
    if (domain === 'info') object = flattenObject(object)

    // update all fields in the object that are in the domain
    const fields = this.fieldsInDomain(domain)
    for (let name in fields) {
      const field = fields[name]
      const value = object[name]
      if (value == null) continue

      const changed = field.updateRange(value)
      if (changed && !quiet) {
        this.sendRangeChangedEvent(field)
      }
    }
  }

  sendRangeChangedEvent (field) {
    if (this._context.eventManager == null) return

    this._context.eventManager.emitSystemEvent('field-range-changed', null, {
      domain: field.domain,
      name: field.name,
      range: field.range
    })
  }
}

// base field class

// The range property is the value of the last batch field range update, not
// the current live value.
class Field {
  constructor (descriptor) {
    this.type = 'unknown'
    this.domain = descriptor.domain
    this.name = descriptor.name
    this.desc = descriptor.desc
    // range set to null means don't keep the range updated, eg agent id
    this.range = descriptor.range ? this.initialRange : null
    this.isEmbedded = this.name.indexOf('.') !== -1
    this.editor = descriptor.editor
    this.shortName = descriptor.shortName
    this.displayName = descriptor.displayName
    this.unit = descriptor.unit
    if (this.isEmbedded) this.path = this.name.split('.')
  }

  // get the value of a field given an agent
  valueFromAgent (agent) {
    if (agent == null) return null

    // special handling for domain "packages"
    if (this.domain === 'packages') {
      switch (this.name) {
        case 'package':
          if (agent.packages == null) return null
          return agent.packages.map(pkg => ({name: pkg.name, version: pkg.version}))
        case 'vulns':
          if (agent.vulns == null) return null
          return agent.vulns.length
        case 'vuln':
          if (agent.vulns == null) return null
          return agent.vulns.map(vuln => vuln.id)
        default:
          return null
      }
    }

    const object = agent[this.domain]
    if (object == null) return null

    if (this.isEmbedded) return getEmbeddedValue(object, this.path)

    return object[this.name]
  }

  // get the initial range of values for this field
  get initialRange () {
    return []
  }

  // anything special that needs to be done before resetting ranges
  prepResetRange () {
  }

  // coerce a value to the valid type for the field
  coerce (value) {
    Logger.warn(`Field subclass did not implement coerce(): ${this}`)
    return value
  }

  // Given a value, update the range for this field.
  // Returns true if the value updated the range, false if it didn't
  updateRange (value) {
    return false
  }
}

// field class for numeric fields
class FieldNumber extends Field {
  constructor (descriptor) {
    super(descriptor)
    this.type = 'number'
  }

  get initialRange () {
    return [0, 0]
  }

  coerce (value) {
    if (typeof value === 'number') return value

    const result = parseFloat(value)
    if (isNaN(result)) return null

    return result
  }

  updateRange (value) {
    // if field doesn't track ranges, don't update
    if (this.range == null) return false

    let range = this.range.slice()

    if (value >= range[0] && value <= range[1]) return false

    if (value < range[0]) range[0] = value
    if (value > range[1]) range[1] = value

    // round range to nice values
    range = nice(range)

    // if no change, return false
    if (!arraysDifferent(range, this.range)) return false

    // if changed, update field range
    this.range = range

    return true
  }
}

// field class for numeric fields
class FieldBoolean extends Field {
  constructor (descriptor) {
    super(descriptor)
    this.type = 'boolean'
  }

  get initialRange () {
    return [true, false]
  }

  coerce (value) {
    if (typeof value === 'boolean') return value

    value = `${value}`.trim()
    return value === 'true'
  }
}

// field class for string fields
class FieldString extends Field {
  constructor (descriptor) {
    super(descriptor)
    this.type = 'string'

    // a parallel range collection to make updates easier
    // defined this way so it's not enumerable during JSON-ization
    Object.defineProperty(this, '_rangeSet', {value: new Set()})
  }

  coerce (value) {
    if (typeof value === 'string') return value

    return `${value}`
  }

  // anything special that needs to be done before resetting ranges
  prepResetRange () {
    this._rangeSet.clear()
  }

  updateRange (value) {
    // if field doesn't track ranges, don't update
    if (this.range == null) return false

    // add values to the set
    let added = false
    const values = Array.isArray(value) ? value : [value]
    for (let aValue of values) {
      if (!this._rangeSet.has(aValue)) {
        added = true
        this._rangeSet.add(aValue)
      }
    }

    // if no new items, no update
    if (!added) return false

    // if new items, recreate range array
    this.range = Array.from(this._rangeSet.keys()).sort()
    return true
  }
}

// field class for packages field
class FieldPackage extends Field {
  constructor (descriptor) {
    super(descriptor)
    this.type = 'package'
  }

  coerce (value) {
    value = `${value}`.trim()

    const pkgVersion = splitPackageVersion(value)
    if (pkgVersion == null) return null
    if (pkgVersion[1] == null) pkgVersion[1] = '*'

    if (!semver.validRange(pkgVersion[1])) return null

    return { name: pkgVersion[0], version: pkgVersion[1] }
  }
}

// match ""@types/foo@1.2.3"
const PackageVersionRegExp1 = /^(@.+?\/.+?)@(.+)$/

// match ""@types/foo"
const PackageVersionRegExp2 = /^(@.+?\/.+)$/

// match "foo@1.2.3"
const PackageVersionRegExp3 = /^(.+?)@(.+)$/

// match for "foo" ; regexp not required as this is the "else" case

// split a string {package}(@{version})? into an array of [package, version]
function splitPackageVersion (string) {
  let match

  match = string.match(PackageVersionRegExp1)
  if (match) return [match[1], match[2]]

  match = string.match(PackageVersionRegExp2)
  if (match) return [match[1], null]

  match = string.match(PackageVersionRegExp3)
  if (match) return [match[1], match[2]]

  return [ string ]
}

// round a numeric range to "nice" numbers
// see: https://github.com/d3/d3-scale/blob/master/README.md#continuous_nice
function nice (range) {
  return d3Scale.scaleLinear()
    .domain(range)
    .nice()
    .domain()
}

// return true if arrays contain different elements
function arraysDifferent (arr1, arr2) {
  if (arr1.length !== arr2.length) return true

  for (let i = 0; i < arr1.length; i++) {
    if (arr1[i] !== arr2[i]) return true
  }

  return false
}

// for nested properties, get the value via the path through the properties
function getEmbeddedValue (object, path) {
  for (let element of path) {
    if (object == null) return null
    object = object[element]
  }

  return object
}

// flatten an object, returning keys with dots for embedded objects
function flattenObject (object, result, prefix) {
  // result and prefix only expected to be set when called recursively
  if (result == null) result = {}
  if (prefix == null) prefix = ''

  for (let propertyName in object) {
    const value = object[propertyName]
    const type = typeof value
    if (value == null || Array.isArray(value) || type === 'string' || type === 'number') {
      result[`${prefix}${propertyName}`] = value
      continue
    }

    flattenObject(value, result, `${prefix}${propertyName}.`)
  }

  return result
}
