'use strict'

exports.handleRequest = handleRequest

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

// Handle the HTTP request
function handleRequest (req, res) {
  const context = req.nsConsole.context
  const qData = req.nsConsole.qData

  const commandParams = {
    command: 'packages',
    id: qData.id
  }

  const agent = context.agentManager.byId(qData.id)
  if (agent == null) {
    return res.status(404).send({ error: {
      message: `process with id ${qData.id} not found`
    }})
  }

  const revokePackagesAdded = agent.revocableOnce('packages-added', () => {
    const vulns = context.vulnManager.getLinkedVulns(agent.packages)
    const packageInfo = context.packageManager.getPackageInfo(agent.packagesHash)

    for (let vuln of vulns) {
      const depsChains = packageInfo.depsChainsMap.get(vuln.id) || []
      vuln.depsChains = depsChains
      vuln.topLevel = depsChains.length
    }

    const responseBody = {
      id: agent.id,
      time: new Date(),
      packages: sortPackages(agent.packages),
      vulnerabilities: sortVulns(vulns)
    }
    res.send(JSON.stringify(responseBody) + '\n')
    endCommand()
  })

  const cmdDispatcher = context.commandDispatcher
  const requestId = cmdDispatcher.send(commandParams)

  const revokeEnd = cmdDispatcher.revocableOnce(`end:${requestId}`, () => {
    Logger.warn('Unexpected end of command', commandParams)
    res.status(500).send('Unexpected end of command.')
    endCommand()
  })

  const revokeError = cmdDispatcher.revocableOnce(`error:${requestId}`, err => {
    Logger.error('Error running command', err, commandParams)
    res.status(500).send(err.message)
    endCommand()
  })

  function endCommand () {
    revokeEnd()
    revokeError()
    revokePackagesAdded()
  }
}

function sortPackages (packages) {
  packages.sort(comparePackages)

  for (let pkg of packages) {
    pkg.dependencies.sort()
    pkg.dependents.sort()
    pkg.vulns.sort()
  }

  return packages
}

function sortVulns (vulns) {
  vulns.sort(compareVulns)

  for (let vuln of vulns) {
    vuln.packages.sort()
  }

  return vulns
}

// package sort function
function comparePackages (a, b) { return a.path.localeCompare(b.path) }

// package sort function
function compareVulns (a, b) { return a.id.localeCompare(b.id) }
