'use strict'

exports.handleRequest = handleRequest

const fs = require('fs')
const path = require('path')
const zlib = require('zlib')
const moment = require('moment')

const util = require('../../util')
const { initializeFlamegraph, baseColors } = require('../../shared')
const svgTemplate = require('../../svg-template')
const { extractOutline, filterData } = require('@ns-private/cpu-profile-massager')
const Logger = util.getLogger(__dirname, __filename)

// Handle the HTTP request
function handleRequest (req, res) {
  const context = req.nsConsole.context
  const config = context.config
  const qData = req.nsConsole.qData
  const diskAssetManager = context.diskAssetManager

  const assetPath = path.join(config.data.dir, 'assets')
  const asset = diskAssetManager.byId(qData.id)

  if (asset == null) {
    res.status(404).send('asset not found')
    return
  }

  if (asset.type !== 'profile') {
    res.status(400).send('asset type is invalid')
    return
  }

  let fileName = path.join(assetPath, asset.fileName)

  const sendSVG = (err, data) => {
    if (err) {
      Logger.error(err, `error processing the ${fileName} - ${err.message}`)
      return res.status(500).send('Internal server error')
    }

    let profileData
    try {
      profileData = JSON.parse(data)
    } catch (error) {
      Logger.error(err, `error parsing the data for ${fileName} - ${err.message}`)
      return res.status(500).send('Internal server error')
    }

    const extractedOutline = extractOutline(profileData)
    const filteredAsset = filterData(profileData, {
      showNSolid: false,
      showProgram: false,
      showGarbage: false,
      showIdle: false
    })

    Logger.debug(`Extracted outline: ${Object.keys(extractedOutline)}`)
    Logger.debug(`Extracted filteredAsset: ${Object.keys(filteredAsset.head)}`)

    const nodes = initializeFlamegraph(filteredAsset.head, () => null, 1160, 580, 20, 200)
    const formattedTime = moment(asset.time).format('MM/DD/YYYY, hh:mm:SS a')
    const svgdata = svgTemplate(asset.pid, asset.hostname, formattedTime, asset.duration / 1000, nodes, baseColors)

    res.set({
      'Content-Type': 'image/svg+xml',
      'Content-Disposition': `attachment; filename=${fileName}.svg`
    })
    res.send(svgdata)
  }

  if (asset.compressed) {
    Logger.debug(`Uncompressing asset ${qData.id}: ${fileName}`)
    gUnzipAsset(`${fileName}.gz`, sendSVG)
  } else {
    fs.readFile(fileName, 'utf-8', sendSVG)
  }
}

function gUnzipAsset (fileName, cb) {
  const buffer = []

  // create unzip stream
  const unzipStream = zlib.createGunzip()
  unzipStream.on('error', err => {
    Logger.error(err, `error unzipping ${fileName} - ${err.message}`)
    return cb(err)
  })

  unzipStream.on('data', data => buffer.push(data.toString()))
  unzipStream.on('end', () => cb(null, buffer.join('')))

  // create the file stream
  const fileStream = fs.createReadStream(fileName)
  fileStream.on('error', err => cb(err))

  fileStream.pipe(unzipStream)
}
