'use strict'

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

const { errors } = require('../../errors')

exports.handleRequest = handleRequest

// Handle the HTTP request
function handleRequest (req, res) {
  const { context } = req.nsConsole
  const { samlToken } = req.body

  // Database login
  //
  //  { "email": "", "password": "" }
  //    returns
  //  200 { "status": "Authorized" }
  //    or
  //  401 { "status": "Unauthorized" }
  //
  // in this case the login is processed immediately

  // Social signin
  //
  //  { "source": "google" }
  //    returns
  //  { "url": "..." }
  //
  //  { "source": "github" }
  //    returns
  //  { "url": "..." }

  if (samlToken) {
    let tokenObj
    try {
      tokenObj = context.provider.verifyToken(samlToken)
    } catch (err) {
      Logger.debug(`SAML session verification failure: ${err}`)
      res.status(400).send({ status: 'Invalid SAML Token' })
      return
    }

    context.provider.setLogin(tokenObj, (err, details) => {
      if (err) {
        Logger.error(`Error registering console: ${err}`)
        if (err instanceof errors.TermsNotAccepted) {
          res.status(400).send({ status: 'Unauthorized', code: 'User::TermsNotAccepted' })
        } else {
          res.status(400).send({ status: 'Unauthorized' })
        }
        return
      }

      const { samlOrgId } = tokenObj
      const keys = context.provider.keysFromDetails(details) || []
      const { key } = keys.find(key => key.orgId === samlOrgId) || {}

      context.provider.syncLicense({ key, notify: true }, (err) => {
        if (err) {
          res.status(400)
          res.send({ error: { message: 'Could not update license.' } })
          Logger.error(err, 'Could not update license.')
          return
        }
        res.status(200)
        res.send(context.provider.license)
      })
    })
    return
  }

  if (req.body.session) {
    if (req.body.registerConsole) {
      // Console cannot be registered with a user session
      res.status(400).send({ status: 'Invalid login method' })
      return
    }

    // User has previously obtained a session from accounts, we should verify
    // it and send it back as a console cookie
    const session = req.body.session
    let permitted
    try {
      permitted = context.provider.userPermitted(session)
    } catch (err) {
      Logger.error(`Error logging in user: ${err}`)
    }
    if (permitted) {
      res.cookie('nsolid-session', session, {
        httpOnly: true,
        expires: new Date(Date.now() + 60 * 60 * 24 * 1000)
      })

      const userRole = permitted
      res.send({ status: 'Authorized', userRole })
    } else {
      res.status(400).send({ status: 'Unauthorized' })
    }
  } else if (req.body.email && req.body.password) {
    let role = req.body.registerConsole ? 'console' : 'user'
    context.provider.login(req.body.email, req.body.password, role, (err, result) => {
      if (err) {
        let errReply
        if (err.code === 'Auth::MfaRequired') {
          errReply = {
            status: 'MfaRequired',
            code: err.code,
            id: err.id,
            session: err.session
          }
        } else {
          Logger.error(`Error logging in: ${err.message}`)
          errReply = {
            status: 'Unauthorized',
            code: err.code
          }
        }
        res.status(400).send(errReply)
      } else {
        if (req.body.registerConsole) {
          context.provider.setLogin(result, (err, details) => {
            if (err) {
              Logger.error(`Error registering console: ${err}`)
              if (err instanceof errors.TermsNotAccepted) {
                res.status(400).send({ status: 'Unauthorized', code: 'User::TermsNotAccepted' })
              } else {
                res.status(400).send({ status: 'Unauthorized' })
              }
              return
            }
            const keys = context.provider.keysFromDetails(details)
            res.send({ keys })
          })
        } else {
          const { session, refreshToken } = result
          res.cookie('nsolid-session', session, {
            httpOnly: true,
            expires: new Date(Date.now() + 60 * 60 * 24 * 1000)
          })
          res.cookie('nsolid-refresh', refreshToken, {
            httpOnly: true,
            expires: new Date(Date.now() + 60 * 60 * 24 * 1000)
          })
          let userRole
          try {
            [ userRole ] = context.provider.userPermitted(session)
          } catch (err) {
            Logger.warn('Could not identify user role for session')
          }
          res.send({ status: 'Authorized', userRole })
        }
      }
    })
  } else if (req.body.mfaCode) {
    context.provider.mfaResponse(req.body.mfaCode, req.body.mfaId, req.body.mfaSession, (err, result) => {
      if (err) {
        Logger.error(`Error logging in: ${err.message}`)
        const errReply = {
          status: 'Unauthorized',
          code: err.code
        }
        res.status(400).send(errReply)
      } else {
        if (req.body.registerConsole) {
          context.provider.setLogin(result, (err, details) => {
            if (err) {
              Logger.error(`Error registering console: ${err}`)
              if (err instanceof errors.TermsNotAccepted) {
                res.status(400).send({ status: 'Unauthorized', code: 'User::TermsNotAccepted' })
              } else {
                res.status(400).send({ status: 'Unauthorized' })
              }
              return
            }
            const keys = context.provider.keysFromDetails(details)
            res.send({ keys })
          })
        } else {
          const { session, refreshToken } = result
          res.cookie('nsolid-session', session, {
            httpOnly: true,
            expires: new Date(Date.now() + 60 * 60 * 24 * 1000)
          })
          res.cookie('nsolid-refresh', refreshToken, {
            httpOnly: true,
            expires: new Date(Date.now() + 60 * 60 * 24 * 1000)
          })
          let userRole
          try {
            [ userRole ] = context.provider.userPermitted(session)
          } catch (err) {
            Logger.warn('Could not identify user role for session')
          }
          res.send({ status: 'Authorized', userRole })
        }
      }
    })
  } else if (req.body.source) {
    let role = req.body.registerConsole ? 'console' : 'user'
    context.provider.ssoUrl(req.body.source, role, (err, result) => {
      if (err) {
        res.status(400).send({ message: err.message })
      } else {
        if (req.body.registerConsole) {
          // console registration doesn't require nonce to be sent back
          context.provider.waitForSso(result.nonce)
          res.send({ url: result.url })
        } else {
          // The UI needs to long poll for the resulting session, so give it
          // the nonce
          res.send({ url: result.url, nonce: result.nonce })
        }
      }
    })
  } else if (req.body.nonce) {
    context.provider.retrieveSession(req.body.nonce, (err, result) => {
      if (err) {
        res.status(400).send({ message: err.message, code: err.code })
        return
      }
      res.send(result)
    })
  } else {
    Logger.error('Invalid login method requested')
    res.status(400).send({ message: 'Invalid login method' })
  }
}
