'use strict'

exports.handleRequest = handleRequest

const joi = require('@hapi/joi')

const util = require('../../util')
const httpUtils = require('../http-utils')
const Logger = util.getLogger(__dirname, __filename)
const uuid = require('uuid').v4
const KEY = 'integrations'

// Handle the HTTP request
function handleRequest (req, res) {
  Logger.debug('handling request', KEY, req.body)
  if (req.body != null) req.body.id = req.body.id || uuid()

  validate(req.body, err => {
    if (err) {
      res.status(400)
      Logger.warn(`invalid integrations request body: ${err.message}`)
      return res.send({ error: { message: err.message } })
    }

    httpUtils.updateSettingByType(req, res, KEY, err => {
      if (err) Logger.warn(err, 'error updating integrations')
    })
  })
}

function validate (body, cb) {
  const invalid = []

  if (body == null) return cb(new Error('JSON request body required'))

  // basic validation of body
  const result = Schema.validate(body, {abortEarly: false})

  if (result.error != null) {
    for (let detail of result.error.details) {
      const path = detail.path.join('.')
      invalid.push(`${path}: ${detail.message}`)
    }
  }

  if (invalid.length !== 0) {
    return cb(new Error(`errors in request body - ${invalid.join('; ')}`))
  }

  // make sure the action integration ids exist
  if (body.type === 'webhook' || body.type === 'slack' || body.type === 'teams') {
    if (body.url == null) return cb(new Error(`${body.type} integration requires a url property`))
    if (body.name == null) body.name = `${body.type} ${body.url}`
  } else if (body.type === 'pagerduty') {
    if (body.key == null) return cb(new Error(`${body.type} integration requires a key property`))
    if (body.name == null) body.name = `${body.type} ${body.key}`
  } else if (body.type === 'email') {
    if (body.emails == null) return cb(new Error(`${body.type} integration requires an emails property`))
    if (body.name == null) body.name = `${body.type} ${body.emails}`
  } else {
    return cb(new Error(`invalid integration type: ${body.type}`))
  }

  return cb()
}

// schemas to validate request body

const Schema = joi.object({
  id: joi.string().required(),
  type: joi.string().required().valid('slack', 'teams', 'webhook', 'email', 'pagerduty'),
  name: joi.string().optional(),
  url: joi.string().optional(),
  key: joi.string().optional(),
  emails: joi.string().optional()
}).unknown(false)
