'use strict'

exports.handleRequest = handleRequest

const joi = require('@hapi/joi')

const util = require('../../util')
const httpUtils = require('../http-utils')
const Logger = util.getLogger(__dirname, __filename)
const KEY = 'notifications'

// Handle the HTTP request
function handleRequest (req, res) {
  Logger.debug('handling request')

  validate(req, err => {
    if (err) {
      res.status(400)
      Logger.warn(`invalid notifications request body: ${err.message}`)
      return res.send({ error: { message: err.message } })
    }

    httpUtils.updateSettingByType(req, res, KEY, err => {
      if (err) Logger.warn(err, 'error updating notifications')
    })
  })
}

function validate (req, cb) {
  const context = req.nsConsole.context
  const body = req.body
  const invalid = []

  if (body == null) return cb(new Error('JSON request body required'))

  // basic validation of body
  const result = Schema.validate(req.body, {abortEarly: false})

  if (result.error != null) {
    for (let detail of result.error.details) {
      const path = detail.path.join('.')
      invalid.push(`${path}: ${detail.message}`)
    }
  }

  if (invalid.length !== 0) {
    return cb(new Error(`errors in request body - ${invalid.join('; ')}`))
  }

  // make sure the action integration ids exist
  context.settingsManager.getSettings('integrations', (err, integrations) => {
    if (err) return cb(err)

    for (let action of body.actions) {
      if (action.integrationId == null) {
        return cb(new Error(`no integrationId: ${JSON.stringify(action)}`))
      } else if (integrations[action.integrationId] == null) {
        return cb(new Error(`invalid integrationId: ${action.integrationId}`))
      }
    }

    return cb()
  })
}

// schemas to validate request body

const actionSchema = joi.object({
  integrationId: joi.string().required()
}).unknown(false)

const notificationEvents = [
  'nsolid-process-blocked',
  'nsolid-vulnerability-detected',
  'nsolid-profile-generated',
  'nsolid-snapshot-generated',
  'nsolid-agent-exit'
]

const Schema = joi.object({
  event: joi.string().required().valid(...notificationEvents),
  actions: joi.array().required().items(actionSchema),
  threshold: joi.number().optional(),
  onlyExceptions: joi.boolean().optional(),
  exitType: joi.string().optional().valid('all', 'normal', 'abnormal'),
  profileWhileBlocked: joi.boolean().optional(),
  profileMinDuration: joi.number().optional()
}).unknown(false)
