'use strict'

exports.handleRequest = handleRequest

const joi = require('@hapi/joi')

const util = require('../../util')
const httpUtils = require('../http-utils')
const Logger = util.getLogger(__dirname, __filename)
const uuid = require('uuid').v4
const queryExpressions = require('../../query-expressions')
const KEY = 'savedViews'

// Handle the HTTP request
function handleRequest (req, res) {
  Logger.debug('handling request')
  if (req.body != null) req.body.id = req.body.id || uuid()

  validate(req, err => {
    if (err) {
      res.status(400)
      Logger.warn(`invalid saved-view request body: ${err.message}`)
      return res.send({ error: { message: err.message } })
    }

    httpUtils.updateSettingByType(req, res, KEY, err => {
      if (err) Logger.warn(err, 'error updating saved-views')
    })
  })
}

const NonIntegrationActions = ['profile', 'snapshot']

function validate (req, cb) {
  const context = req.nsConsole.context
  const body = req.body
  const invalid = []

  if (body == null) return cb(new Error('JSON request body required'))

  // basic validation of body
  const result = Schema.validate(req.body, {abortEarly: false})

  if (result.error != null) {
    for (let detail of result.error.details) {
      const path = detail.path.join('.')
      invalid.push(`${path}: ${detail.message}`)
    }
  }

  if (invalid.length !== 0) {
    return cb(new Error(`errors in request body - ${invalid.join('; ')}`))
  }

  // make sure the expression parses
  try {
    queryExpressions.fromJSON(context, body.expression)
  } catch (err) {
    return cb(new Error(`invalid expression: ${err.message}`))
  }

  body.id = req.body.id || uuid()
  body.actions = body.actions || []

  // make sure the action integration ids exist
  context.settingsManager.getSettings('integrations', (err, integrations) => {
    if (err) return cb(err)
    for (let action of body.actions) {
      if (action.integrationId != null) {
        if (integrations[action.integrationId] == null) {
          return cb(new Error(`invalid integrationId: ${action.integrationId}`))
        }
      } else if (action.type == null) {
        return cb(new Error(`no action type: ${JSON.stringify(action)}`))
      } else if (NonIntegrationActions.indexOf(action.type) === -1) {
        return cb(new Error(`invalid action type: ${action.type}`))
      }

      if (action.type === 'profile') {
        if (action.duration == null) {
          return cb(new Error(`no duration specified for profile`))
        }
      }
    }

    // make axes are valid fields
    let field
    const fieldManager = context.fieldManager

    field = fieldManager.getField(body.axisX)
    if (field == null) return cb(new Error(`invalid axisX value: ${body.axisX}`))

    field = fieldManager.getField(body.axisY)
    if (field == null) return cb(new Error(`invalid axisY value: ${body.axisY}`))

    return cb()
  })
}

// schemas to validate request body

const filterTermSchema = joi.object({
  field: joi.string().required(),
  operator: joi.string().required(),
  values: joi.array().required()
}).unknown(false)

const filterExprSchema = joi.array().items(filterTermSchema)

const actionSchema = joi.object({
  integrationId: joi.string().optional(),
  type: joi.string().optional(),
  duration: joi.number().optional()
}).unknown(false)

const Schema = joi.object({
  id: joi.string().required(),
  name: joi.string().required(),
  expression: filterExprSchema.required(),
  axisX: joi.string().required(),
  axisY: joi.string().required(),
  scalingX: joi.string().optional(),
  scalingY: joi.string().optional(),
  snooze: joi.number().optional(),
  delay: joi.number().optional(),
  actions: joi.array().optional().items(actionSchema),
  persistent: joi.boolean().optional()
}).unknown(false)
