'use strict'

exports.create = create
exports.createRps = createRps
exports.backfillCQs = backfillCQs

const asynclib = require('async')

const api = require('./api')
const util = require('../util')

const Logger = util.getLogger(__dirname, __filename)

function create (context, cb) {
  asynclib.series([
    next => { createDb(context, next) },
    // 2019-05-14: disabling backfill mode until it can be debugged
    next => { createRps(context, false, next) },
    next => { createCQs(context, next) }
  ], err => {
    cb(err)
  })
}

var policies = [
  {
    name: 'agentData',
    retentionHours: 1,
    shardRetentionHours: 1,
    startupRetentionHours: 24,
    startupShardRetentionHours: 24,
    default: true
  },
  {
    name: 'one_minute_daily',
    retentionDays: 1,
    shardRetentionHours: 1,
    startupShardRetentionHours: 24
  },
  {
    name: 'one_hour_monthly',
    retentionDays: 31,
    shardRetentionDays: 7,
    useSettingsDuration: true
  },
  {
    name: 'rp_events',
    retentionDays: 365,
    shardRetentionDays: 7
  }
]

// FIXME: https://github.com/nodesource/nsolid-console-ng/issues/363
var columnAggTypes = {
  mean: [
    'activeHandles', 'activeRequests', 'cpuPercent', 'cpuSpeed',
    'cpuSystemPercent', 'cpuUserPercent', 'freeMem', 'gcCpuPercent',
    'gcDurUs99Ptile', 'gcDurUsMedian', 'heapSizeLimit', 'heapTotal',
    'heapUsed', 'externalMem', 'load15m', 'load1m', 'load5m', 'loopAvgTasks',
    'loopEstimatedLag', 'loopIdlePercent', 'loopsPerSecond', 'rss',
    'totalAvailableSize', 'totalHeapSizeExecutable', 'totalPhysicalSize',
    'httpClientMedian', 'httpClient99Ptile', 'httpServerMedian',
    'httpServer99Ptile', 'dnsMedian', 'dns99Ptile'
  ],
  max: [
    'gcCount', 'gcForcedCount', 'gcFullCount', 'gcMajorCount', 'systemUptime',
    'uptime', 'blockInputOpCount', 'blockOutputOpCount',
    'ctxSwitchInvoluntaryCount', 'ctxSwitchVoluntaryCount', 'ipcReceivedCount',
    'ipcSentCount', 'loopTotalCount', 'pageFaultHardCount',
    'pageFaultSoftCount', 'signalCount', 'swapCount', 'vulns',
    'httpClientCount', 'httpServerCount', 'dnsCount', 'httpClientAbortCount',
    'httpServerAbortCount'
  ],
  last: ['app', 'hostname', 'id', 'tags', 'title', 'user']
}

const columns = []
for (let func of Object.keys(columnAggTypes)) {
  for (let name of columnAggTypes[func]) {
    columns.push({ name, func })
  }
}

var cqs = [
  {
    name: 'downsample_to_daily',
    columns,
    fromRp: 'agentData',
    toRp: 'one_minute_daily',
    window: '1m',
    tags: ['id', 'app', 'tags']
  },
  {
    name: 'downsample_to_monthly',
    columns,
    fromRp: 'agentData',
    toRp: 'one_hour_monthly',
    window: '1h',
    tags: ['id']
  }
]

function createDb (context, cb) {
  const createCommand = 'CREATE DATABASE nsolid'
  const url = context.config.influxdb.url
  const user = context.config.influxdb.user
  const password = context.config.influxdb.password

  api.query(url, 'nsolid', createCommand, user, password, err => {
    if (err) {
      Logger.error(err, 'error creating database nsolid')
    }
    return cb(err)
  })
}

// On first-pass use a "starting up" mode
// when old data has quiesced, change it to "running" mode
function createRps (context, backfillMode, cb) {
  const getRetention = 'SHOW RETENTION POLICIES ON nsolid'
  const url = context.config.influxdb.url
  const user = context.config.influxdb.user
  const password = context.config.influxdb.password

  api.query(url, 'nsolid', getRetention, user, password, (err, body) => {
    if (err) {
      Logger.error(err, 'error checking existing retention policies')
      return cb(err)
    }

    // The result will look something like this:
    // { "results": [ {
    //   "series": [ {
    //     "columns": [ "name", "duration", "shardGroupDuration", "replicaN", "default" ],
    //     "values": [
    //       [ "default", "0", "168h0m0s", 1, false ],
    //       [ "agentData", "720h0m0s", "24h0m0s", 1, true ]
    //     ]
    //   } ]
    // } ] }

    var currentPolicies
    try {
      const response = JSON.parse(body)
      currentPolicies = response.results[0].series[0].values
    } catch (err) {
      Logger.error(err, 'error checking existing retention policies')
    }

    asynclib.eachSeries(policies, (policy, next) => {
      createOrAlterRP(context, currentPolicies, policy, backfillMode, next)
    }, err => {
      if (err) {
        Logger.error(err, 'error setting replication policies')
      }
      return cb(err)
    })
  })
}

function createOrAlterRP (context, currentPolicies, spec, backfillMode, cb) {
  let exists
  let currentRetentionHours
  let currentShardRetentionHours

  // Only the longest RP will be impacted by the user-configurable
  // NSOLID_CONSOLE_DATA_RETENTION_DAYS
  let requestedRetentionHours = spec.useSettingsDuration
    ? (parseInt(context.config.data.retentionDays, 10) * 24)
    : spec.retentionDays ? (spec.retentionDays * 24)
    : backfillMode && spec.startupRetentionHours ? spec.startupRetentionHours
    : spec.retentionHours

  let requestedShardRetentionHours = spec.shardRetentionDays ? (spec.shardRetentionDays * 24)
    : backfillMode && spec.startupShardRetentionHours ? spec.startupShardRetentionHours
    : spec.shardRetentionHours

  const matching = currentPolicies.filter(pol => { return pol[0] === spec.name })
  if (matching.length) {
    exists = true

    let match = matching[0][1].match(/^(\d+)h.*/)
    if (match) {
      currentRetentionHours = parseInt(match[1], 10)
    }
    match = matching[0][2].match(/^(\d+)h.*/)
    if (match) {
      currentShardRetentionHours = parseInt(match[1], 10)
    }
  }

  if (currentRetentionHours != null) {
    Logger.info(`current retention policy for ${spec.name}: ${currentRetentionHours} hour(s)`)
  }

  if (!exists || (currentRetentionHours !== requestedRetentionHours) || (currentShardRetentionHours !== requestedShardRetentionHours)) {
    Logger.info(`setting retention policy ${spec.name}: ${requestedRetentionHours} hour(s)`)
    const verb = exists ? 'ALTER' : 'CREATE'
    // Do not set a default RP on external influx
    const isDefault = spec.default && !context.usingExternalInflux
    const retentionCommand = `${verb} RETENTION POLICY ${spec.name} ON nsolid ` +
      `DURATION ${requestedRetentionHours}h REPLICATION 1 SHARD DURATION ` +
      `${requestedShardRetentionHours}h ${isDefault ? 'DEFAULT' : ''}`

    const url = context.config.influxdb.url
    const user = context.config.influxdb.user
    const password = context.config.influxdb.password

    api.query(url, 'nsolid', retentionCommand, user, password, cb)
  } else {
    cb()
  }
}

function createCQs (context, cb) {
  asynclib.eachSeries(cqs, (cq, next) => {
    createCQ(context, cq, next)
  }, err => {
    if (err) {
      Logger.error(err, 'error creating continuous queries')
    }
    return cb(err)
  })
}

function createCQ (context, spec, cb) {
  // first, delete the CQ if it exists
  var dropQuery = `drop continuous query ${spec.name} on nsolid`
  const url = context.config.influxdb.url
  const user = context.config.influxdb.user
  const password = context.config.influxdb.password

  api.query(url, 'nsolid', dropQuery, user, password, err => {
    if (err) {
      Logger.error(err, `error replacing existing continuous query ${spec.name}`)
      return cb(err)
    }

    var createQuery = `CREATE CONTINUOUS QUERY ${spec.name} ON nsolid BEGIN SELECT ` +
      spec.columns.map(column => { return `${column.func}("${column.name}") as "${column.name}"` }).join(', ') +
      ` INTO "${spec.toRp}"."metrics" ` +
      ` FROM "${spec.fromRp}"."metrics" ` +
      ` GROUP BY time(${spec.window}), ` +
      spec.tags.join(', ') +
      ' END'

    api.query(url, 'nsolid', createQuery, user, password, err => {
      if (err) {
        Logger.error(err, `error creating continuous query ${spec.name}`)
        return cb(err)
      }
      cb()
    })
  })
}

// This takes millisecond timestamps and runs all CQs as backfill for the
// periods covering them.
function backfillCQs (context, startTimestamp, endTimestamp, cb) {
  const url = context.config.influxdb.url
  const user = context.config.influxdb.user
  const password = context.config.influxdb.password

  asynclib.eachSeries(cqs, (spec, next) => {
    const backfillQuery = 'SELECT ' +
      spec.columns.map(column => { return `${column.func}("${column.name}") as "${column.name}"` }).join(', ') +
      `INTO "${spec.toRp}"."metrics" ` +
      `FROM "${spec.fromRp}"."metrics" ` +
      `WHERE time >= ${String(startTimestamp) + '000000'} AND time <= ${String(endTimestamp) + '000000'} ` +
      `GROUP BY time(${spec.window}), ` +
      spec.tags.join(', ')

    api.query(url, 'nsolid', backfillQuery, user, password, err => {
      if (err) {
        Logger.error(err, `error backfilling continuous query ${spec.name}`)
        return next(err)
      }
      next()
    })
  }, err => {
    if (err) {
      Logger.error(err, 'error backfilling continuous queries')
    }
    return cb(err)
  })
}
