'use strict'

const lodash = require('lodash')

const api = require('./api')
const util = require('../util')

const createRps = require('./schema').createRps
const backfillCQs = require('./schema').backfillCQs

exports.create = create

const SkipEventLoggingTypes = new Set([
  'field-range-changed',
  'influxdb-error'
])

const Logger = util.getLogger(__dirname, __filename)

// Create and return a new database writer.
function create (context) {
  return new Writer(context)
}

class Writer {
  constructor (context) {
    this._context = context
    this._writerMetrics = new QueuedWriter(context, 'agentData')
    this._writerEvents = new QueuedWriter(context, 'rp_events')
    this.ready = true

    // 2019-05-14: disabling backfill mode until it can be debugged
    this.rpMode = 'standard'
    // this.oldDataStart = Date.now()
    // this.lastWroteOldData = null
    // this.updateResetRpTimer()
  }

  // Write an metrics record to the database.
  writeMetrics (agent, message) {
    // Ensure sane shape of body before modifying it
    const data = lodash.defaults({}, message.body)

    message.body = data
    const line = this._createLine(agent, 'metrics', message)
    this._writerMetrics.writeLine(line)
  }

  // Write a system event record to the database.
  writeEvent (name, agentId, args, time) {
    if (SkipEventLoggingTypes.has(name)) return

    const timestamp = util.uniqDateNS(time)

    const agentIdPart = (agentId == null)
      ? ''
      : `agent=${this.escapeFieldValue(agentId)},`

    if (args === undefined) args = null
    const bodyPart = `body=${this.escapeFieldValue(args)}`
    const typePart = `type=${this.escapeKey(name)}`

    const line = `events,${typePart} ${agentIdPart}${bodyPart} ${timestamp}`
    this._writerEvents.writeLine(line)
  }

  // influx keys and tags must escape "=", " ", and ","
  escapeKey (key) {
    return key.replace(/([ =,])/g, '\\$1')
  }

  // format influx field values
  escapeFieldValue (value) {
    if (value == null) return '"null"'  // return null as the string 'null'
    if (lodash.isNumber(value)) return value
    if (lodash.isBoolean(value)) return `${value}`

    // if it smells like a string, ensure it's a string
    if (lodash.isString(value)) {
      value = `${value}`
    // last chance - it's an object, so JSON stringify
    } else {
      value = JSON.stringify(value)
    }

    // return quoted string, escaping quotes
    return `"${value.replace(/"/g, '\\"')}"`
  }

  _createLine (agent, measurement, message) {
    const tags = Object.keys(agent.tags).length
      ? ',' + lodash.map(agent.tags, (val, key) => `${this.escapeKey(key)}=${this.escapeKey(val)}`).join(',')
      : ''

    const fields = lodash.map(message.body, (val, key) => `${this.escapeKey(key)}=${this.escapeFieldValue(val)}`).join(',')

    // 2019-05-14: disabling backfill mode until it can be debugged
    // const timeMS = this.messageIsOld(message)
    // if (timeMS) {
    //   this.lastWroteOldData = Date.now()
    //   if (timeMS < this.oldDataStart) {
    //     this.oldDataStart = timeMS
    //   }

    //   // Bump the reset timer since we've seen new backfill data
    //   if (this.rpMode === 'backfill') {
    //     this.updateResetRpTimer()
    //   }
    // }
    const timestamp = message.timeNS || util.uniqDateNS()
    return `${measurement}${tags} ${fields} ${timestamp}`
  }

  updateResetRpTimer () {
    clearTimeout(this.resetRpTimer)
    this.resetRpTimer = setTimeout(() => {
      this.standardMode()
    }, this._context.config.data.backfillTimeoutMs).unref()
  }

  messageIsOld (message) {
    if (message.timeNS) {
      const timeMS = parseInt(message.timeNS.slice(0, -6), 10)
      if (timeMS < Date.now() - this._context.config.data.oldDataThresholdMs) {
        return timeMS
      }
    }
    // Without a timestamp we can't know
    return false
  }

  // The backfillMode and standardMode functions are async but dont require
  // callbacks as they're using the ready property as a semaphore.  The agent
  // objects will queue metrics if ready is not true
  // 2019-05-14: disabling backfill mode until it can be debugged
  // backfillMode () {
  //   this.rpMode = 'modifying'
  //   this.ready = false
  //
  //   Logger.info('Switching to backfill mode')
  //   createRps(this._context, true, err => {
  //     if (err) {
  //       Logger.error(`Could not modify RPs to backfill mode: ${err}`)
  //     }
  //     this.ready = true
  //     this.rpMode = 'backfill'
  //     // Set a high-water mark for this backfill cycle
  //     this.oldDataStart = Date.now()
  //     this.updateResetRpTimer()
  //   })
  // }

  standardMode () {
    this.resetRpTimer = null
    this.rpMode = 'modifying'
    this.ready = false

    Logger.info('Data stream stable, adjusting RPs to standard mode')
    // Go back an extra minute on the CQ backfill
    backfillCQs(this._context, this.oldDataStart - 60000, this.lastWroteOldData, err => {
      if (err) {
        Logger.error(`Error backfilling CQs: ${err}`)
      }
      createRps(this._context, false, err => {
        if (err) {
          Logger.error(`Could not modify RPs to standard mode: ${err}`)
        }
        this.ready = true
        this.rpMode = 'standard'
      })
    })
  }
}

class QueuedWriter {
  constructor (context, rp) {
    this._context = context
    this._rp = rp
    this._user = context.config.influxdb.user !== '' ? context.config.influxdb.user : null
    this._password = context.config.influxdb.password !== '' ? context.config.influxdb.password : null
    this._timeout = null
    this._queue = []
    this._writing = false
    this._warnedQueueTruncation = 0
    this._errorState = false

    // Buffer writes for this many millis.
    this.BUFFER_FOR_MS = this._context.config.influxdb.bufferForMs || 250
    this.CHUNK_SIZE = this._context.config.influxdb.chunkSize || 100
    this.MAX_QUEUE_LENGTH = this._context.config.influxdb.maxQueueLength || 100000
  }

  writeLine (line) {
    Logger.debug(`queuing line to InfluxDB: ${line}`)
    this._queue.push(line)

    if (this._queue.length >= this.CHUNK_SIZE) {
      this._writeQueued()
      return
    }

    if (this._timeout) return

    this._timeout = setTimeout(() => this._writeQueued(), this.BUFFER_FOR_MS)
    this._timeout.unref()
  }

  _writeQueued () {
    // Do this synchronously because data can arrive while influx is writing
    // and we want to allow a new timer to start in the meantime.
    if (this._timeout) clearTimeout(this._timeout)
    this._timeout = null

    if (this._queue.length === 0) {
      if (this._warnedQueueTruncation) {
        this._warnedQueueTruncation = 0
        if (this._context.eventManager) {
          this._context.eventManager.emitSystemEvent('influxdb-recovered', null, { message: 'write recovery' })
        }
      }
      return
    }

    // Do not allow the queue to grow unbounded
    if (this._queue.length > this.MAX_QUEUE_LENGTH) {
      // Allow this warning to be shown once per minute
      const now = Date.now()
      if (!this._warnedQueueTruncation || this._warnedQueueTruncation < now - 60000) {
        this._warnedQueueTruncation = now
        Logger.warn('Truncating metrics because Influx write queue is full')
        if (this._context.eventManager) {
          this._context.eventManager.emitSystemEvent('influxdb-truncated', null, { message: 'write queue truncated' })
        }
      }
      this._queue = this._queue.slice(this._queue.length - this.MAX_QUEUE_LENGTH)
    }

    // Don't try to write while Influx is still launching.  This assumes that
    // influxdb will be fully running and the db initialized by the time
    // influxdbProcess is saved to Context.
    if (!this._context.influxdbProcess && !this._context.usingExternalInflux) {
      return
    }

    // Only write one chunk at a time.  If an extended buffering period has
    // been in place, there could be many thousands of lines to process, so
    // don't stampede.
    if (this._writing) return
    this._writing = true

    Logger.debug(`writing ${this._queue.length} queued line to InfluxDB`)

    // Write the first 100 lines in the queue
    const joinedLines = this._queue.slice(0, this.CHUNK_SIZE).join('\n')
    this._queue = this._queue.slice(this.CHUNK_SIZE)

    api.write(this._context.config.influxdb.url, 'nsolid', joinedLines, this._rp, this._user, this._password, err => {
      this._writing = false
      if (err) {
        // TODO: consider requeuing data if using external influx
        Logger.error(err, 'error while writing metrics to Influx, data discarded')
        if (this._context.eventManager) {
          this._context.eventManager.emitSystemEvent('influxdb-error', null, { message: 'write error' })
        }
        this._errorState = true
        return
      }

      if (this._errorState) {
        if (this._context.eventManager) {
          this._context.eventManager.emitSystemEvent('influxdb-recovered', null, { message: 'write recovery' })
        }
        this._errorState = false
      }
      if (this._queue.length >= this.CHUNK_SIZE) {
        this._writeQueued()
      }
    })
  }
}
