'use strict'

// this module creates a package manager which manages packages and vulns
// for a process based on a hash of the process's packages payload from the agent

exports.create = create

const path = require('path')
const crypto = require('crypto')

const DepsChains = require('./deps-chains')
const RevocableEventEmitter = require('./RevocableEventEmitter')

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)

// number of minutes between garbage collecting unused packages
const GC_INTERVAL_MINS = 5

function create (context) {
  return new PackageManager(context)
}

class PackageManager extends RevocableEventEmitter {
  constructor (context) {
    super()

    this._context = context
    this._cachedPackageInfo = new Map()

    const gcInterval = setInterval(() => this._gc(), GC_INTERVAL_MINS * 60 * 1000)
    gcInterval.unref()

    // when the vulndb has been updated, update all the package vulns
    context.eventManager.on('vulnerabilities-database-updated', () => {
      this.updateAllPackageVulns()
      this._context.eventManager.emitSystemEvent('package-vulnerabilities-updated')
    })
  }

  // return a hash of the packages data
  // side-effect: if packages not already cached, cache them
  hashPackages (packagesPayload) {
    if (packagesPayload == null) packagesPayload = []
    const normalizedPackages = normalizePackagesPayload(packagesPayload)

    const hash = crypto.createHash('sha1')
    hash.update(JSON.stringify(normalizedPackages))
    const packagesHash = hash.digest('hex')

    // if we've already seen this set of packages, return hash
    if (this._cachedPackageInfo.has(packagesHash)) return packagesHash

    // otherwise, calculate dependents, vulns, then cache packages and vulns
    addDependents(normalizedPackages)

    // create, update, and cache a packageInfo object
    const packageInfo = {
      packages: normalizedPackages,
      vulns: null,
      depsChainsMap: null
    }

    this.updatePackageVulns(packageInfo)
    this._cachedPackageInfo.set(packagesHash, packageInfo)

    return packagesHash
  }

  // update one package info structure with vulns
  updatePackageVulns (packageInfo) {
    const packages = packageInfo.packages

    packageInfo.vulns = this._context.vulnManager.setVulnsForPackages(packages)
    packageInfo.depsChainsMap = calculateDepsChainsMap(packages, packageInfo.vulns)
  }

  // update all the packages vuln data, used when vuln db updated
  updateAllPackageVulns () {
    for (let packageInfo of this._cachedPackageInfo.values()) {
      this.updatePackageVulns(packageInfo)
    }
  }

  getPackageInfo (hash) {
    return this._cachedPackageInfo.get(hash)
  }

  getPackages (hash) {
    const packageInfo = this.getPackageInfo(hash)
    if (packageInfo == null) return null

    return packageInfo.packages
  }

  getVulns (hash) {
    const packageInfo = this.getPackageInfo(hash)
    if (packageInfo == null) return null

    return packageInfo.vulns
  }

  getDepsChainsMap (hash) {
    const packageInfo = this.getPackageInfo(hash)
    if (packageInfo == null) return null

    return packageInfo.depsChainsMap
  }

  // clean up unused cached packages and vulns
  _gc () {
    const agents = this._context.agentManager.list()
    const packagesToDelete = new Set(this._cachedPackageInfo.keys())

    // delete the hash for all agents in the set of packages to delete
    for (let agent of agents) {
      packagesToDelete.delete(agent.packagesHash)
    }

    // packagesToDelete now contains packages to delete
    for (let hash of packagesToDelete) {
      this._cachedPackageInfo.delete(hash)
    }
  }
}

function calculateDepsChainsMap (packages, vulns) {
  return DepsChains.getDepsChainsMap(packages, vulns)
}

function addDependents (packages) {
  // for each pkg, normalize dependencies and dependents arrays, add to map
  const pkgMap = new Map()
  for (let pkg of packages) {
    if (pkg.dependencies == null) pkg.dependencies = []
    pkg.dependents = []
    pkgMap.set(pkg.path, pkg)
  }

  // for every pkg's dependency, add the pkg as a dependent
  //   "path": "/path/to/node_modules/arr-diff",
  //   "dependencies": [
  //    "../arr-flatten",
  for (let pkg of packages) {
    for (let depPkgRelPath of pkg.dependencies) {
      const depPkgPath = path.resolve(pkg.path, depPkgRelPath)
      const depPkg = pkgMap.get(depPkgPath)
      if (depPkg == null) {
        Logger.warn(`can't find package ${depPkgPath} from ${pkg.path}`)
        continue
      }

      depPkg.dependents.push(path.relative(depPkgPath, pkg.path))
    }
  }
}

function normalizePackagesPayload (packagesPayload) {
  const packages = packagesPayload.map(packageElement => ({
    name: packageElement.name,
    version: packageElement.version,
    path: packageElement.path,
    main: packageElement.main,
    dependencies: normalizeDependencies(packageElement.dependencies),
    flagFiles: normalizeFlagFiles(packageElement.flagFiles)
  }))

  // sort packages by path (which includes package name)
  packages.sort((a, b) => a.path.localeCompare(b.path))

  return packages
}

function normalizeDependencies (dependencies) {
  if (dependencies == null) return

  return dependencies.slice().sort()
}

// Normalize the names of the snyk flag files.
// Used to contain colons, now using hyphens.
function normalizeFlagFiles (flagFiles) {
  if (flagFiles == null) return

  flagFiles = flagFiles.map(flagFile => flagFile.replace(/:/g, '-'))
  flagFiles.sort()

  return flagFiles
}
