'use strict'

exports.evaluate = evaluateWrapped

const lodash = require('lodash')
const semver = require('semver')

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)

function evaluateWrapped (term, agentValue) {
  try {
    return evaluate(term, agentValue)
  } catch (err) {
    Logger.warn(err, `error evaluating term "${term}"`)
    return false
  }
}

// Evaluate a term against an agent's value(s); returns true or false
function evaluate (term, agentValue) {
  // packages require special treatment
  if (term.field.type === 'package') {
    return evaluatePackage(term, agentValue)
  }

  // for agent values that are arrays (eg tags),
  // if any element value evals true, return true; else return false
  if (Array.isArray(agentValue)) {
    for (let subAgentValue of agentValue) {
      const result = evaluate(term, subAgentValue)
      if (result) return true
    }
    return false
  }

  // terms may also have multiple values
  // if any element value evals true, return true; else return false
  for (let termValue of term.values) {
    const result = evaluateBasic(term, termValue, agentValue)
    if (result) return true
  }

  return false
}

function evaluateBasic (term, termValue, agentValue) {
  switch (term.operator) {
    case '=': return agentValue === termValue
    case '!=': return agentValue !== termValue
    case '<': return agentValue < termValue
    case '<=': return agentValue <= termValue
    case '>': return agentValue > termValue
    case '>=': return agentValue >= termValue
    case '~': return wildcardMatch(termValue, agentValue)
    case '!~': return !wildcardMatch(termValue, agentValue)
  }

  Logger.warn(`invalid operator in query term: ${term.operator}`)
  return false
}

function wildcardMatch (termValue, agentValue) {
  // create a string form of a regexp, escaping all the bits, and then
  // converting the '*' chars into '.*'
  const frontAnchor = (/^\*/.test(termValue) ? '' : '^')
  const rearAnchor = (/\*$/.test(termValue) ? '' : '$')
  const termPattern = termValue
    .split(/\*+/g)
    .map(lodash.escapeRegExp)
    .join('.*')
  const regexpString = frontAnchor + termPattern + rearAnchor

  const regexp = new RegExp(regexpString)
  return regexp.test(agentValue)
}

function evaluatePackage (term, agentValue) {
  const operator = term.operator
  const termValue = term.values[0]

  const matchingPackages = getMatchingPackages(termValue.name, agentValue)

  switch (operator) {
    case '=': return evalPackageEq(termValue, matchingPackages)
    case '!=': return evalPackageNeq(termValue, matchingPackages)
    case '<': return evalPackageLt(termValue, matchingPackages)
    case '<=': return evalPackageLte(termValue, matchingPackages)
    case '>': return evalPackageGt(termValue, matchingPackages)
    case '>=': return evalPackageGte(termValue, matchingPackages)
    default: return false
  }
}

function evalPackageEq (termValue, matchingPackages) {
  if (termValue.version === '*') return matchingPackages.length > 0

  if (!semver.validRange(termValue.version)) return false

  for (let pkg of matchingPackages) {
    if (semver.valid(pkg.version)) {
      if (semver.satisfies(pkg.version, termValue.version)) return true
    }
  }

  return false
}

function evalPackageNeq (termValue, matchingPackages) {
  if (termValue.version === '*') return matchingPackages.length === 0

  if (!semver.validRange(termValue.version)) return false

  for (let pkg of matchingPackages) {
    if (semver.valid(pkg.version)) {
      if (semver.satisfies(pkg.version, termValue.version)) return false
    }
  }

  return true
}

function evalPackageLt (termValue, matchingPackages) {
  return evalPackagRelational('<', termValue, matchingPackages)
}

function evalPackageLte (termValue, matchingPackages) {
  return evalPackagRelational('<=', termValue, matchingPackages)
}

function evalPackageGt (termValue, matchingPackages) {
  return evalPackagRelational('>', termValue, matchingPackages)
}

function evalPackageGte (termValue, matchingPackages) {
  return evalPackagRelational('>=', termValue, matchingPackages)
}

function evalPackagRelational (operator, termValue, matchingPackages) {
  const semverRange = `${operator}${termValue.version}`
  if (!semver.validRange(semverRange)) return false

  for (let pkg of matchingPackages) {
    if (semver.valid(pkg.version)) {
      if (semver.satisfies(pkg.version, semverRange)) return true
    }
  }

  return false
}

function getMatchingPackages (packageName, packages) {
  if (packages == null) return []
  return packages.filter(pkg => pkg.name === packageName)
}
