'use strict'

const fs = require('fs')
const URL = require('url')
const path = require('path')
const crypto = require('crypto')

const toml = require('toml')
const lodash = require('lodash')

const logger = require('./logger')

exports.getLogger = getLogger
exports.printHelp = printHelp
exports.randomId = randomId
exports.sequencer = sequencer
exports.cbNextTick = cbNextTick
exports.onlyCallOnce = onlyCallOnce
exports.fileExists = fileExists
exports.ensureDirExists = ensureDirExists
exports.loadFile = loadFile
exports.parseAddress = parseAddress
exports.normalizeAddress = normalizeAddress
exports.dateMS2ISO = dateMS2ISO
exports.uniqDateNS = uniqDateNS
exports.parseIntStrict = parseIntStrict
exports.requireZmq = requireZmq
exports.readSchema = readSchema
exports.getHttpAgent = getHttpAgent
exports.getHttpBaseURL = getHttpBaseURL
exports.getProjectPath = getProjectPath
exports.createRPromise = createRPromise

// loaded late because of mutual deps ...
const config = require('./config')

// Return a logger for the given file name.
function getLogger (names) {
  return logger.getLogger.apply(logger, arguments)
}

const Logger = getLogger(__dirname, __filename)

// Print the CLI help.
function printHelp () {
  const helpFile = path.join(__dirname, '..', 'etc', 'help.md')
  const helpText = fs.readFileSync(helpFile, 'utf8')
  console.log(helpText.trim().replace('$sample-config-file$', config.SampleConfigFile))

  const defEnvVars = lodash.map(
    config.getDefaultEnvVars(),
    (val, key) => [key, val]
  )

  const keyLen = lodash.max(lodash.map(defEnvVars, (el) => el[0].length))

  console.log('')
  console.log('Default configuration as environment variables:')
  console.log('')

  console.log(padRight('environment variable', keyLen), ' ', 'value')

  const sep = padRight('', keyLen).replace(/ /g, '-')
  console.log(sep, ' ', sep)
  for (let defEnvVar of defEnvVars) {
    console.log(padRight(defEnvVar[0], keyLen), ' ', defEnvVar[1])
  }
}

// Pad a string on the right with spaces.
function padRight (s, len) {
  s = `${s}`
  while (s.length < len) s += ' '
  return s
}

// Run a callback with the specified args on the next tick.
function cbNextTick (cb, args) {
  args = [].slice.call(arguments, 1)
  setImmediate(() => cb.apply(null, args))
}

// Return a random id of hex characters
function randomId (len) {
  len = len || 20
  return crypto.randomBytes(len).toString('hex')
}

// Return a function that returns sequence numbers.
function sequencer () {
  let curr = 0

  return function sequenced () {
    curr++

    if (curr >= Number.MAX_SAFE_INTEGER) curr = 1
    return curr
  }
}

// Create a version of a function which will only be called once
function onlyCallOnce (fn) {
  let called = false

  return function onlyCalledOnce () {
    if (called) return
    called = true

    return fn.apply(null, arguments)
  }
}

// Return indication of whether file exists.
function fileExists (file) {
  let stats
  try {
    stats = fs.statSync(file)
  } catch (err) {
    return false
  }

  return stats.isFile()
}

// Ensure that the specified directory exists, and if not, create it.
function ensureDirExists (dirName) {
  try {
    fs.mkdirSync(dirName, { recursive: true })
  } catch (err) {
    Logger.error(err, `error creating directory: ${dirName}`)
    return err
  }

  return null
}

// Load the specified file as a string.
function loadFile (file) {
  try {
    return fs.readFileSync(file, 'utf8')
  } catch (err) {
    return null
  }
}

// Parse a normalized address.
// The address should be in hostname:port form already.
// Returns [hostname, port].
function parseAddress (address) {
  if (!address) return null

  const match = address.match(/(.*):(.*)/)
  if (!match) return null

  return [match[1], match[2]]
}

// Normalize an address to hostname:port.
function normalizeAddress (address, defaultPort) {
  let hostname = ''
  let port = ''
  let protocol

  // peel off http:// or https:// protocol
  const match = address.match(/^(https?:\/\/)(.*)$/)
  if (match) {
    protocol = match[1]
    address = match[2]
  }

  // if a hostname or port; no `:` allowed in address without a port at end
  if (address.indexOf(':') === -1) {
    // port
    if (isInteger(address)) {
      port = address
    // hostname
    } else {
      hostname = address
    }
  // has a `:` so parse as hostname:port
  } else {
    const parts = parseAddress(address)

    hostname = parts[0]
    port = parts[1]
  }

  // apply default values
  if (hostname === '') hostname = '0.0.0.0'
  if (port === '') port = `${defaultPort}`

  if (protocol) hostname = `${protocol}${hostname}`

  if (isNaN(parseInt(port, 10))) {
    throw new Error(`non-numeric port specified in address: '${address}'`)
  }

  return `${hostname}:${port}`
}

// Return indication of whether the value is integral, even if it's a string.
function isInteger (value) {
  value = `${value}`
  return value.match(/^\d+$/) != null
}

// Convert a date in unix epoch ms to an ISO string.
function dateMS2ISO (ms) {
  return new Date(ms).toISOString()
}

let UniqDateSeq = 0
let UniqDatePrefix = '00000'

// Return the date in nanos, each one unique.
function uniqDateNS (date) {
  if (date == null || !(String(date).match(/^\d{13}$/))) {
    date = Date.now()
  }
  const result = `${date}${UniqDatePrefix}${UniqDateSeq}`

  UniqDateSeq++
  switch (UniqDateSeq) {
    case 10: UniqDatePrefix = '0000'; break
    case 100: UniqDatePrefix = '000'; break
    case 1000: UniqDatePrefix = '00'; break
    case 10000: UniqDatePrefix = '0'; break
    case 100000: UniqDatePrefix = ''; break
    case 1000000:
      UniqDateSeq = 0
      UniqDatePrefix = '00000'
  }

  return result
}

// Parse an integer, strictly.
function parseIntStrict (string) {
  if (string == null) return Number.NaN

  string = `${string}`
  if (string.match(/^-?\d+$/ == null)) return Number.NaN

  return parseInt(string, 10)
}

// Do a require on zmq - try N|Solid's version, then npm version if not there.
function requireZmq () {
  let mod

  try {
    mod = require('__nsolid_zmq')
    Logger.debug('loaded ZeroMQ via __nsolid_zmq')
    return mod
  } catch (err) {
    Logger.debug('N|Solid module _zmq not found, trying npm version')
  }

  Logger.warn('unable to load built-in zmq from nsolid, attempting to load from npm module for diagnostic purposes')

  try {
    mod = require('zmq')
    Logger.debug('loaded ZeroMQ via zmq')
    return mod
  } catch (err) {
    Logger.error('unable to load zmq module')
    throw err
  }
}

function readSchema (name) {
  const fileName = path.join(__dirname, '..', 'schemas', `${name}.toml`)
  const schemaToml = loadFile(fileName)
  if (schemaToml == null) return null

  let schema
  try {
    schema = toml.parse(schemaToml)
  } catch (err) {
    if (err.name === 'SyntaxError') {
      const location = `${fileName}:${err.line}:${err.column}`
      Logger.error(err, `error parsing toml file ${location}: ${err.message}`)
    } else {
      Logger.error(err, 'error parsing toml file')
    }
    schema = null
  }

  return schema
}

// values for host names in server addresses that mean "wildcard", and should
// be base-valued at localhost
const WildcardHosts = [
  '0.0.0.0',
  '[::]',
  '*'
]

// Return the base url for this server; eg, http://localhost:6753/
function getHttpBaseURL (context) {
  const config = context.config
  const proto = config.web.proto
  const addr = config.web.server

  const match = addr.match(/^(.*?):(\d+)$/)
  if (match == null) return null

  let host = match[1]
  const port = match[2]

  if (WildcardHosts.indexOf(host) !== -1) host = 'localhost'

  return `${proto}://${host}:${port}`
}

// Return the appropriate http agent for the specified URL
function getHttpAgent (context, url) {
  const parsedURL = URL.parse(url)

  if (parsedURL.protocol === 'http:') return context.httpAgent
  if (parsedURL.protocol === 'https:') return context.httpsAgent

  return null
}

// given a filename, get it's project-relative path
function getProjectPath (fileName) {
  return path.relative(path.join(__dirname, '..'), fileName)
}

// return a promise that has explicit resolve and reject methods
function createRPromise () {
  let resolver
  let rejecter

  const promise = new Promise((resolve, reject) => {
    resolver = resolve
    rejecter = reject
  })

  promise.resolve = (val) => resolver(val)
  promise.reject = (val) => rejecter(val)

  return promise
}

/**
 * Expose a Map which is serialized as an array of its values.
 */
class SerializableMap extends Map {
  toString () {
    return Array.from(this.values())
  }

  toJSON () {
    return Array.from(this.values())
  }
}

exports.SerializableMap = SerializableMap
