'use strict'

const walkTree = require('./walk-tree')

const getTimeByNode = (data) => {
  const timeByNode = {}

  if (data.timestamps) {
    for (let i = 0; i < data.samples.length; i++) {
      if (!timeByNode[i]) {
        timeByNode[i] = 0
      }

      // CDT calculates this as the "time before" the sample rather than the
      // "time after", so we will also use this for consistency.  For the first
      // node, assume the standard interval of 1ms
      timeByNode[i] += i ? data.timestamps[i] - data.timestamps[i - 1] : 1000
    }
  }

  // timestamps is in microseconds, so normalize to milliseconds
  for (let key of Object.keys(timeByNode)) {
    timeByNode[key] = timeByNode[key] / 1000
  }

  return timeByNode
}

const extractOutline = (data) => {
  // compute this based on both samples and timestamps.  Older runtimes may not
  // produce timestamps, but if present they are more accurate.
  let nonIdleHits = 0
  let nonIdleTime = 0
  let gcHits = 0
  let gcTime = 0
  let programHits = 0
  let programTime = 0

  const timeByNode = getTimeByNode(data)

  walkTree(data.head, node => {
    node.selfTime = (timeByNode[node.id] || 0) * node.hitCount

    // do not include these special counts into nonIdleTime
    if (node.functionName === '(idle)') {
      return false
    }

    if (node.functionName === '(garbage collector)') {
      gcHits += node.hitCount
      gcTime += node.selfTime
      return false
    }

    if (node.functionName === '(program)') {
      programHits += node.hitCount
      programTime += node.selfTime
      return false
    }

    nonIdleHits += node.hitCount
    nonIdleTime += node.selfTime
  })

  const totalHits = data.samples.length
  const totalTime = data.timestamps ? (data.timestamps[data.timestamps.length - 1] - data.timestamps[0]) / 1000 : 0

  const program = programTime ? (programTime / totalTime) : (programHits / totalHits)
  const gc = gcTime ? (gcTime / totalTime) : (gcHits / totalHits)
  const application = nonIdleTime ? (nonIdleTime / totalTime) : (nonIdleHits / totalHits)

  // samples that weren't bucketed will be added to idle
  const idle = 1 - (program + gc + application)

  return { idle, program, gc, application }
}

module.exports = extractOutline
