'use strict'

const path = require('path')
const lodash = require('lodash')
const walkTree = require('./walk-tree')

const filterByUrl = (node, parent, filterUrlString) => {
  if (parent && node.url === filterUrlString) {
    const index = parent.children.findIndex(item => item === node)
    for (let i = 0; i < node.children.length; i++) {
      parent.children.push(node.children[i])
    }
    if (index > -1) {
      parent.children.splice(index, 1)
    }
    return true
  }

  for (let j = 0; j < node.children.length; j++) {
    if (filterByUrl(node.children[j], node, filterUrlString)) {
      return true
    }
  }
}

// modifies the CPU profile with data required in this view (which is then
// passed onto the visualizations). this should be done whenever the new data
// comes in, or different filtering options are chosen
//
// also ensures that leaves have a nonzero hitCount, fixing an unwanted
// side-effect of calling filterByUrl
const traverseProfile = (data) => {
  let nodeCount = 0
  const _walk = (node, parent) => {
    // ensure that leaves have a minimum hitCount of 1
    if (node.hitCount === 0 && node.children.length === 0) {
      node.hitCount = 1
    }
    let totalHits = node.hitCount
    let totalTime = node.selfTime

    // recursively count the total number of hits
    // of all child nodes for each node
    for (let i = 0; i < node.children.length; i++) {
      const ret = _walk(node.children[i], node)
      totalHits += ret[0]
      totalTime += ret[1]
    }

    // core code - no path on script name
    // user code, path in script name does not include node_modules
    // 3rd-party code, path in script name includes node_modules
    const coreCode = node.url.indexOf('/') === -1 || node.url.startsWith('internal/')
    const userCode = node.url.indexOf('node_modules') === -1

    const short = userCode
      ? [
        path.basename(path.dirname(node.url)),
        path.basename(node.url)
      ].filter(Boolean).join('/').replace(/^\.\//g, '')
      : node.url.replace(/^.+?ode_modules\/(.*)$/, '$1')

    node.parent = parent
    node.totalHitCount = totalHits
    node.totalTime = totalTime
    node.short = short
    node.codeType = coreCode ? 'core' : (userCode ? 'user' : ('thirdParty'))

    nodeCount++

    return [totalHits, totalTime]
  }
  _walk(data.head, null)
  data.head.nodeCount = nodeCount
}

// traverses the CPU profile tree, associating a unique
// numerical index for each node based on their callUID.
// This is used for highlighting multiple occurrences of
// the same function on hover.

// @return {Object} An object index where each key is
//                  the callUID and each value is the
//                  numerical index.
const generateUIDs = (tree) => {
  let uidCount = 0
  let uidList = {}
  walkTree(tree, node => {
    uidList[node.callUID] = uidList[node.callUID] || ++uidCount
  })
  return uidList
}

const filterData = (data, dataOptions) => {
  let filteredData = lodash.cloneDeep(data)
  filteredData.head.children = data.head.children
    .filter(node => (
      (node.functionName !== '(program)' || dataOptions.showProgram) &&
      (node.functionName !== '(idle)' || dataOptions.showIdle) &&
      (node.functionName !== '(garbage collector)' || dataOptions.showGarbage)
    ))
  if (!dataOptions.showNSolid) {
    while (filterByUrl(filteredData.head, null, 'nsolid.js'));
  }
  traverseProfile(filteredData) // do this last
  filteredData.UIDs = generateUIDs(filteredData.head)
  return filteredData
}

module.exports = filterData
