'use strict'

const fs = require('fs')
const tape = require('tape')
const almostEqual = require('almost-equal')

const { extractOutline, filterData, searchTree } = require('../index')
const { sumOutline } = require('./util')
const largeParsedAsset = JSON.parse(fs.readFileSync('./test/fixtures/large.cpuprofile'))

tape('check a few nodes in a large profile', (t) => {
  const extractedOutline = extractOutline(largeParsedAsset)
  const filteredAsset = filterData(largeParsedAsset, {
    showNSolid: false,
    showProgram: false,
    showGarbage: false,
    showIdle: false
  })

  const expectedOutline = {
    idle: 0.8674689517865655,
    program: 0.011818986706066697,
    gc: 0.014465450640378186,
    application: 0.10624661086698965
  }
  t.ok(almostEqual(sumOutline(extractedOutline), 1, almostEqual.FLT_EPSILON, almostEqual.FLT_EPSILON), 'outline adds up to 1.0')
  t.deepEquals(extractedOutline, expectedOutline)

  t.equals(filteredAsset.typeId, 'CPU')
  t.equals(filteredAsset.uid, '3')
  t.equals(filteredAsset.title, 'undefined')
  t.equals(filteredAsset.startTime, 19792948)
  t.equals(filteredAsset.endTime, 19792997)
  t.equals(filteredAsset.samples.length, 45112)
  t.equals(filteredAsset.timestamps.length, 45112)
  t.equals(Object.keys(filteredAsset.UIDs).length, 866)

  const root = filteredAsset.head
  t.equals(root.functionName, '(root)')
  t.equals(root.url, '')
  t.equals(root.lineNumber, 0)
  t.equals(root.bailoutReason, '')
  t.equals(root.id, 1)
  t.equals(root.scriptId, 0)
  t.equals(root.hitCount, 0)
  t.equals(root.children.length, 18)

  const parserOnHeadersComplete = root.children[0]
  t.equals(parserOnHeadersComplete.functionName, 'parserOnHeadersComplete')
  t.equals(parserOnHeadersComplete.url, '_http_common.js')
  t.equals(parserOnHeadersComplete.lineNumber, 45)
  t.equals(parserOnHeadersComplete.callUID, 209)
  t.equals(parserOnHeadersComplete.bailoutReason, 'no reason')
  t.equals(parserOnHeadersComplete.id, 6)
  t.equals(parserOnHeadersComplete.scriptId, 85)
  t.equals(parserOnHeadersComplete.hitCount, 1)
  t.equals(parserOnHeadersComplete.children.length, 3)
  t.equals(parserOnHeadersComplete.lineTicks.length, 1)
  t.equals(parserOnHeadersComplete.selfTime, 1.034)
  t.equals(parserOnHeadersComplete.parent, root)
  t.equals(parserOnHeadersComplete.totalHitCount, 372)
  t.equals(parserOnHeadersComplete.totalTime, 399.8279999999999)
  t.equals(parserOnHeadersComplete.short, '_http_common.js')
  t.equals(parserOnHeadersComplete.codeType, 'core')

  const _tickCallback = root.children[1]
  t.equals(_tickCallback.functionName, '_tickCallback')
  t.equals(_tickCallback.url, 'internal/process/next_tick.js')
  t.equals(_tickCallback.lineNumber, 93)
  t.equals(_tickCallback.callUID, 720)
  t.equals(_tickCallback.bailoutReason, 'no reason')
  t.equals(_tickCallback.id, 39)
  t.equals(_tickCallback.scriptId, 45)
  t.equals(_tickCallback.hitCount, 11)
  t.equals(_tickCallback.children.length, 3)
  t.equals(_tickCallback.lineTicks.length, 1)
  t.equals(_tickCallback.selfTime, 12.144000000000002)
  t.equals(_tickCallback.parent, root)
  t.equals(_tickCallback.totalHitCount, 3784)
  t.equals(_tickCallback.totalTime, 4043.1669999999995)
  t.equals(_tickCallback.short, 'process/next_tick.js')
  t.equals(_tickCallback.codeType, 'core')

  const runMicrotasks = _tickCallback.children[1]
  t.equals(runMicrotasks.functionName, 'runMicrotasks')
  t.equals(runMicrotasks.url, '')
  t.equals(runMicrotasks.lineNumber, 0)
  t.equals(runMicrotasks.callUID, 718)
  t.equals(runMicrotasks.bailoutReason, '')
  t.equals(runMicrotasks.id, 180)
  t.equals(runMicrotasks.scriptId, 0)
  t.equals(runMicrotasks.hitCount, 15)
  t.equals(runMicrotasks.children.length, 2)
  t.equals(runMicrotasks.lineTicks.length, 1)
  t.equals(runMicrotasks.selfTime, 16.125)
  t.equals(runMicrotasks.parent, _tickCallback)
  t.equals(runMicrotasks.totalHitCount, 3347)
  t.equals(runMicrotasks.totalTime, 3570.735999999999)
  t.equals(runMicrotasks.short, '.')
  t.equals(runMicrotasks.codeType, 'core')

  const processImmediate = root.children[2]
  t.equals(processImmediate.functionName, 'processImmediate')
  t.equals(processImmediate.url, 'timers.js')
  t.equals(processImmediate.lineNumber, 600)
  t.equals(processImmediate.callUID, 793)
  t.equals(processImmediate.bailoutReason, 'no reason')
  t.equals(processImmediate.id, 121)
  t.equals(processImmediate.scriptId, 40)
  t.equals(processImmediate.hitCount, 0)
  t.equals(processImmediate.children.length, 1)
  t.equals(processImmediate.lineTicks, undefined)
  t.equals(processImmediate.selfTime, 0)
  t.equals(processImmediate.parent, root)
  t.equals(processImmediate.totalHitCount, 380)
  t.equals(processImmediate.totalTime, 408.636)
  t.equals(processImmediate.short, 'timers.js')
  t.equals(processImmediate.codeType, 'core')

  // drill down to into an anonymous function marked as 'thirdParty'
  const g = processImmediate.children[0].children[0].children[0].children[0].children[0]
  const anonymous = g.children[0]
  t.equals(anonymous.functionName, '')
  t.equals(anonymous.url, '/apps/nodejs/sl-personalization/node_modules/request/request.js')
  t.equals(anonymous.lineNumber, 520)
  t.equals(anonymous.callUID, 790)
  t.equals(anonymous.bailoutReason, 'no reason')
  t.equals(anonymous.id, 127)
  t.equals(anonymous.scriptId, 338)
  t.equals(anonymous.hitCount, 1)
  t.equals(anonymous.children.length, 1)
  t.deepEquals(anonymous.lineTicks, [{ hitCount: 1, line: 520 }])
  t.equals(anonymous.selfTime, 1.086)
  t.equals(anonymous.parent, g)
  t.equals(anonymous.totalHitCount, 372)
  t.equals(anonymous.totalTime, 400.0450000000001)
  t.equals(anonymous.short, 'request/request.js')
  t.equals(anonymous.codeType, 'thirdParty')

  t.end()
})

tape('test searchTree with a large profile', t => {
  const filteredAsset = filterData(largeParsedAsset, {
    showNSolid: false,
    showProgram: false,
    showGarbage: false,
    showIdle: false
  })

  const expected = [{
    id: 1217,
    codeType: 'user',
    functionName: 'a.addTaggingHeaderinOptions',
    hitCount: 3,
    selfTime: 3.216,
    short: 'nodeagent/nodejsagent.js',
    totalHitCount: 11,
    totalTime: 11.568000000000001,
    url: '/int/dyn/dynatrace/agent/res/nodeagent/nodejsagent.js',
    lineNumber: 3682
  }]

  const searchResults = searchTree(filteredAsset.head, 'erin')
  t.deepEquals(searchResults, expected, 'found Erin')

  t.end()
})

tape('test with showIdle: true', t => {
  const filteredAsset = filterData(largeParsedAsset, {
    showNSolid: false,
    showProgram: false,
    showGarbage: false,
    showIdle: true
  })

  const root = filteredAsset.head
  t.equals(root.functionName, '(root)', 'functionName')
  t.equals(root.url, '')
  t.equals(root.lineNumber, 0)
  t.equals(root.bailoutReason, '')
  t.equals(root.id, 1)
  t.equals(root.scriptId, 0)
  t.equals(root.hitCount, 0)
  t.equals(root.children.length, 19)

  const idle = root.children[0]
  t.equals(idle.functionName, '(idle)', 'functionName')
  t.equals(idle.url, '')
  t.equals(idle.lineNumber, 0)
  t.equals(idle.callUID, 4)
  t.equals(idle.bailoutReason, '')
  t.equals(idle.id, 5)
  t.equals(idle.scriptId, 0)
  t.equals(idle.hitCount, 39477)
  t.equals(idle.children.length, 0)
  t.equals(idle.lineTicks, undefined)
  t.equals(idle.selfTime, 42319.344000000005)
  t.equals(idle.parent, root)
  t.equals(idle.totalHitCount, 39477)
  t.equals(idle.totalTime, 42319.344000000005)
  t.equals(idle.short, '.')
  t.equals(idle.codeType, 'core')

  t.end()
})

tape('test with showGarbage: true', t => {
  const filteredAsset = filterData(largeParsedAsset, {
    showNSolid: false,
    showProgram: false,
    showGarbage: true,
    showIdle: false
  })

  const root = filteredAsset.head
  t.equals(root.functionName, '(root)')
  t.equals(root.url, '')
  t.equals(root.lineNumber, 0)
  t.equals(root.bailoutReason, '')
  t.equals(root.id, 1)
  t.equals(root.scriptId, 0)
  t.equals(root.hitCount, 0)
  t.equals(root.children.length, 19)

  const garbageCollector = root.children[7]
  t.equals(garbageCollector.functionName, '(garbage collector)', 'functionName')
  t.equals(garbageCollector.url, '', 'url')
  t.equals(garbageCollector.lineNumber, 0)
  t.equals(garbageCollector.callUID, 853)
  t.equals(garbageCollector.bailoutReason, '')
  t.equals(garbageCollector.id, 646)
  t.equals(garbageCollector.scriptId, 0)
  t.equals(garbageCollector.hitCount, 653)
  t.equals(garbageCollector.children.length, 0)
  t.equals(garbageCollector.lineTicks, undefined)
  t.equals(garbageCollector.selfTime, 707.8520000000001)
  t.equals(garbageCollector.parent, root)
  t.equals(garbageCollector.totalHitCount, 653)
  t.equals(garbageCollector.totalTime, 707.8520000000001)
  t.equals(garbageCollector.short, '.')
  t.equals(garbageCollector.codeType, 'core')

  t.end()
})

tape('test with showProgram: true', t => {
  const filteredAsset = filterData(largeParsedAsset, {
    showNSolid: false,
    showProgram: true,
    showGarbage: false,
    showIdle: false
  })

  const root = filteredAsset.head
  t.equals(root.functionName, '(root)')
  t.equals(root.url, '')
  t.equals(root.lineNumber, 0)
  t.equals(root.bailoutReason, '')
  t.equals(root.id, 1)
  t.equals(root.scriptId, 0)
  t.equals(root.hitCount, 0)
  t.equals(root.children.length, 19)

  const program = root.children[2]
  t.equals(program.functionName, '(program)', 'functionName')
  t.equals(program.url, '', 'url')
  t.equals(program.lineNumber, 0)
  t.equals(program.callUID, 721)
  t.equals(program.bailoutReason, '')
  t.equals(program.id, 120)
  t.equals(program.scriptId, 0)
  t.equals(program.hitCount, 538)
  t.equals(program.children.length, 0)
  t.equals(program.lineTicks, undefined)
  t.equals(program.selfTime, 578.35)
  t.equals(program.parent, root)
  t.equals(program.totalHitCount, 538)
  t.equals(program.totalTime, 578.35)
  t.equals(program.short, '.')
  t.equals(program.codeType, 'core')

  t.end()
})

tape('test with showNSolid: true', t => {
  const filteredAsset = filterData(largeParsedAsset, {
    showNSolid: true,
    showProgram: false,
    showGarbage: false,
    showIdle: false
  })

  const root = filteredAsset.head
  t.equals(root.functionName, '(root)')
  t.equals(root.url, '')
  t.equals(root.lineNumber, 0)
  t.equals(root.bailoutReason, '')
  t.equals(root.id, 1)
  t.equals(root.scriptId, 0)
  t.equals(root.hitCount, 0)
  t.equals(root.children.length, 14)

  const program = root.children[9]
  t.equals(program.functionName, 'metricsCommand', 'functionName')
  t.equals(program.url, 'nsolid.js', 'url')
  t.equals(program.lineNumber, 950)
  t.equals(program.callUID, 343)
  t.equals(program.bailoutReason, 'TryCatchStatement')
  t.equals(program.id, 811)
  t.equals(program.scriptId, 67)
  t.equals(program.hitCount, 1)
  t.equals(program.children.length, 2)
  t.deepEquals(program.lineTicks, [{ hitCount: 1, line: 955 }])
  t.equals(program.selfTime, 1.054)
  t.equals(program.parent, root)
  t.equals(program.totalHitCount, 29)
  t.equals(program.totalTime, 30.904)
  t.equals(program.short, 'nsolid.js')
  t.equals(program.codeType, 'core')

  t.end()
})
