"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const AdjacentMapFileResolver_1 = require("./AdjacentMapFileResolver");
const RemoteFileResolver_1 = require("./RemoteFileResolver");
const SourceInfo_1 = require("./SourceInfo");
const DEFAULT_MAX_COLUMN_TRIES = 40;
class SourceInfoMapError {
    constructor(failureReason, errorMessage) {
        this.failureReason = failureReason;
        this.errorMessage = errorMessage;
    }
}
exports.SourceInfoMapError = SourceInfoMapError;
class SourceInfoMapResult {
    constructor(sourceInfo, failureReason = null, errorMessage = null) {
        this.sourceInfo = sourceInfo;
        this.failureReason = failureReason;
        this.errorMessage = errorMessage;
    }
    get hasError() { return this.failureReason != null; }
    get isEmpty() { return this.sourceInfo == null; }
    static failureResult(reason, errorMessage) {
        return new SourceInfoMapResult(null, reason, errorMessage);
    }
    static success(sourceInfo) {
        return new SourceInfoMapResult(sourceInfo);
    }
    static get empty() {
        return new SourceInfoMapResult(null);
    }
}
exports.SourceInfoMapResult = SourceInfoMapResult;
class SourceInfoMapper {
    constructor(_sourcemapResolver, _maxColumnTries) {
        this._sourcemapResolver = _sourcemapResolver;
        this._maxColumnTries = _maxColumnTries;
    }
    static forAdjacentFile(maxColumnTries = DEFAULT_MAX_COLUMN_TRIES) {
        const sourceMapResolver = new AdjacentMapFileResolver_1.AdjacentMapFileResolver();
        return new SourceInfoMapper(sourceMapResolver, maxColumnTries);
    }
    static forRemoteFile(endpoint, zipMapFile, generateRequestInfo = RemoteFileResolver_1.defaultGenerateRequestInfo, maxColumnTries = DEFAULT_MAX_COLUMN_TRIES) {
        const sourceMapResolver = new RemoteFileResolver_1.RemoteFileResolver(endpoint, zipMapFile, generateRequestInfo);
        return new SourceInfoMapper(sourceMapResolver, maxColumnTries);
    }
    static forResolver(sourcemapResolver, maxColumnTries = DEFAULT_MAX_COLUMN_TRIES) {
        return new SourceInfoMapper(sourcemapResolver, maxColumnTries);
    }
    async map(nodes) {
        const errors = [];
        for (const node of nodes) {
            const children = node.children;
            const res = await this._map(node.url, SourceInfo_1.SourceInfo.fromNode(node));
            if (res.hasError) {
                errors.push(new SourceInfoMapError(res.failureReason, res.errorMessage));
            }
            if (!res.isEmpty) {
                const { url, line, functionName } = res.sourceInfo;
                node.url = url;
                node.lineNumber = line;
                node.functionName = functionName || '';
            }
            if (children != null && children.length > 0) {
                const newErrors = await this.map(children);
                newErrors.forEach(err => errors.push(err));
            }
        }
        return errors;
    }
    async _map(originalFile, generated) {
        const res = await this._sourcemapResolver.resolveSourcemap(originalFile);
        if (res.hasError) {
            return SourceInfoMapResult.failureResult(res.failureReason, res.errorMessage);
        }
        const sourcemap = res.sourcemap;
        if (sourcemap == null)
            return SourceInfoMapResult.empty;
        for (let column = 0; column < this._maxColumnTries; column++) {
            const orig = await sourcemap.originalPosition(generated.withColumn(column));
            if (orig.line !== 0 || orig.column !== 0) {
                if (orig.functionName == null) {
                    orig.functionName = generated.functionName;
                }
                if (originalFile.length > 0 && orig.url.length > 0) {
                    orig.url = path_1.default.resolve(path_1.default.dirname(originalFile), orig.url);
                }
                return SourceInfoMapResult.success(orig);
            }
        }
        return SourceInfoMapResult.empty;
    }
}
exports.SourceInfoMapper = SourceInfoMapper;
//# sourceMappingURL=SourceInfoMapper.js.map