import { dirname } from 'path'
import { canRead, isAccessibleDirectory, readFile } from './fs-utils'
import { isuserFile } from './path-utils'
import { Sourcemap } from './Sourcemap'
import { ResolveSourcemapResult, SourcemapResolver } from './SourcemapResolver'

export class ProjectInaccessibleError extends Error {
  constructor(msg: string) {
    super(msg)
    this.name = ProjectInaccessibleError.Name
  }

  static get Name(): string { return 'ProjectInaccessibleError' }
}

async function canAccessProject(url: string): Promise<boolean> {
  const canAccessSourceFile = await canRead(url)
  if (canAccessSourceFile) return true
  const directory = dirname(url)
  return isAccessibleDirectory(directory)
}

export class AdjacentMapFileResolver extends SourcemapResolver {
  constructor(private verifyProjectAccess: boolean = false) { super() }

  async resolveSourcemap(url: string): Promise<ResolveSourcemapResult> {
    if (!isuserFile(url)) return ResolveSourcemapResult.empty

    if (this.verifyProjectAccess) {
      const haveAccess = await canAccessProject(url)
      if (!haveAccess) {
        throw new ProjectInaccessibleError(`Cannot access project for ${url}.`)
      }
    }
    const json = await this._readMapFile(url)
    if (json == null) return ResolveSourcemapResult.empty
    return new ResolveSourcemapResult(
      Sourcemap.fromRawSourcemapJSON(json)
    )
  }

  private _mapFilePath(generatedFile: string) {
    return `${generatedFile}.map`
  }

  private async _readMapFile(generatedFile: string): Promise<string | null> {
    const p = this._mapFilePath(generatedFile)
    const readable = await canRead(p)
    if (!readable) return null
    return readFile(p, 'utf8')
  }
}
