import fetch from 'cross-fetch'
import { RawSourceMap } from 'source-map'
import { isuserFile } from './path-utils'
import { Sourcemap } from './Sourcemap'
import {
  ResolveSourcemapFailureReason,
  ResolveSourcemapResult,
  SourcemapResolver
} from './SourcemapResolver'

async function projectAccessible(res: Response): Promise<boolean> {
  try {
    const body = await res.json()
    return !body.projectInaccessible
  } catch (err) {
    return false
  }
}

type GenerateRequestInfo =
  (fileUrl: string, zipFile: string | null) => RequestInit

export const defaultGenerateRequestInfo: GenerateRequestInfo =
  (fileUrl, zipMapFile) => {
    const payload = zipMapFile != null ? { fileUrl, zipMapFile } : { fileUrl }
    return {
      method: 'POST',
      credentials: 'include',
      mode: 'cors',
      headers: {
        'Content-Type': 'application/json',
        'Accept': 'application/json'
      },
      body: JSON.stringify(payload)
    } as RequestInit
  }

export class RemoteFileResolver extends SourcemapResolver {
  constructor(
    private _endpoint: string,
    private _zipMapFile: string | null = null,
    private _generateRequestInfo: GenerateRequestInfo =
      defaultGenerateRequestInfo,
    cache: boolean = true) {
    super(cache)
  }

  async resolveSourcemap(fileUrl: string): Promise<ResolveSourcemapResult> {
    if (!isuserFile(fileUrl)) return ResolveSourcemapResult.empty

    const requestInfo = this._generateRequestInfo(fileUrl, this._zipMapFile)
    try {
      const res = await fetch(this._endpoint, requestInfo)
      if (res.status === 404) {
        return await projectAccessible(res)
          ? ResolveSourcemapResult.empty
          : ResolveSourcemapResult.failureResult(
            ResolveSourcemapFailureReason.projectInaccessible,
            `Cannot access the project containing ${fileUrl}`
          )
      }
      if (!res.ok) return ResolveSourcemapResult.empty
      const body = await res.json()
      if (body.error != null) throw new Error(body.error)
      if (body.result == null) return ResolveSourcemapResult.empty
      return new ResolveSourcemapResult(
        Sourcemap.fromRawSourcemap(body.result as RawSourceMap)
      )
    } catch (err) {
      return ResolveSourcemapResult.failureResult(
        ResolveSourcemapFailureReason.unknown,
        err.toString()
      )
    }
  }

}
