import path from 'path'
import { AdjacentMapFileResolver } from './AdjacentMapFileResolver'
import { CpuProfileNode } from './CpuProfile'
import {
  defaultGenerateRequestInfo,
  RemoteFileResolver
} from './RemoteFileResolver'
import { SourceInfo } from './SourceInfo'
import {
  ResolveSourcemapFailureReason,
  SourcemapResolver
} from './SourcemapResolver'

const DEFAULT_MAX_COLUMN_TRIES = 40

export class SourceInfoMapError {
  constructor(
    public failureReason: ResolveSourcemapFailureReason,
    public errorMessage: string
  ) { }
}

export class SourceInfoMapResult {
  get hasError(): boolean { return this.failureReason != null }

  get isEmpty(): boolean { return this.sourceInfo == null }

  static failureResult(
    reason: ResolveSourcemapFailureReason,
    errorMessage: string): SourceInfoMapResult {
    return new SourceInfoMapResult(null, reason, errorMessage)
  }

  static success(sourceInfo: SourceInfo): SourceInfoMapResult {
    return new SourceInfoMapResult(sourceInfo)
  }

  static get empty(): SourceInfoMapResult {
    return new SourceInfoMapResult(null)
  }

  constructor(
    public sourceInfo: SourceInfo | null,
    public failureReason: ResolveSourcemapFailureReason | null = null,
    public errorMessage: string | null = null
  ) { }
}

export class SourceInfoMapper {
  static forAdjacentFile(maxColumnTries = DEFAULT_MAX_COLUMN_TRIES) {
    const sourceMapResolver = new AdjacentMapFileResolver()
    return new SourceInfoMapper(sourceMapResolver, maxColumnTries)
  }

  static forRemoteFile(
    endpoint: string,
    zipMapFile: string,
    generateRequestInfo = defaultGenerateRequestInfo,
    maxColumnTries = DEFAULT_MAX_COLUMN_TRIES
  ): SourceInfoMapper {
    const sourceMapResolver =
      new RemoteFileResolver(endpoint, zipMapFile, generateRequestInfo)
    return new SourceInfoMapper(sourceMapResolver, maxColumnTries)
  }

  static forResolver(
    sourcemapResolver: SourcemapResolver,
    maxColumnTries = DEFAULT_MAX_COLUMN_TRIES
  ) {
    return new SourceInfoMapper(sourcemapResolver, maxColumnTries)
  }

  private constructor(
    private _sourcemapResolver: SourcemapResolver,
    private _maxColumnTries: number) { }

  async map(nodes: CpuProfileNode[]): Promise<SourceInfoMapError[]> {
    const errors: SourceInfoMapError[] = []

    for (const node of nodes) {
      const children = node.children
      const res = await this._map(node.url, SourceInfo.fromNode(node))
      if (res.hasError) {
        errors.push(
          new SourceInfoMapError(res.failureReason!, res.errorMessage!)
        )
      }
      if (!res.isEmpty) {
        const { url, line, functionName } = res.sourceInfo!
        node.url = url
        node.lineNumber = line
        node.functionName = functionName || ''
      }
      if (children != null && children.length > 0) {
        const newErrors = await this.map(children)
        newErrors.forEach(err => errors.push(err))
      }
    }
    return errors
  }

  private async _map(
    originalFile: string,
    generated: SourceInfo
  ): Promise<SourceInfoMapResult> {
    const res =
      await this._sourcemapResolver.resolveSourcemap(originalFile)
    if (res.hasError) {
      return SourceInfoMapResult.failureResult(
        res.failureReason!, res.errorMessage!
      )
    }
    const sourcemap = res.sourcemap
    if (sourcemap == null) return SourceInfoMapResult.empty

    for (let column = 0; column < this._maxColumnTries; column++) {
      const orig =
        await sourcemap.originalPosition(generated.withColumn(column))
      if (orig.line !== 0 || orig.column !== 0) {
        if (orig.functionName == null) {
          orig.functionName = generated.functionName
        }
        if (originalFile.length > 0 && orig.url.length > 0) {
          orig.url = path.resolve(path.dirname(originalFile), orig.url)
        }
        return SourceInfoMapResult.success(orig)
      }
    }
    return SourceInfoMapResult.empty
  }
}
