import JSZip, { JSZipObject, loadAsync } from 'jszip'

import assert = require('assert')
import { readFile } from './fs-utils'
import { Sourcemap } from './Sourcemap'
import { ResolveSourcemapResult, SourcemapResolver } from './SourcemapResolver'

export type MapFileFromPath = (path: string) => string

export class ZipMapFileResolver extends SourcemapResolver {
  static adjacentMapFileFromPath: MapFileFromPath = path => `${path}.map`

  static cachedInstance(
    zipFilePath: string,
    mapFileFromPath?: MapFileFromPath
  ): ZipMapFileResolver {
    if (ZipMapFileResolver._resolverCache.has(zipFilePath)) {
      return ZipMapFileResolver._resolverCache.get(zipFilePath)!
    }
    const resolver = new ZipMapFileResolver(zipFilePath, mapFileFromPath)
    ZipMapFileResolver._resolverCache.set(zipFilePath, resolver)
    return resolver
  }

  static removeFromCache(zipFilePath: string) {
    if (!ZipMapFileResolver._resolverCache.has(zipFilePath)) return
    ZipMapFileResolver._resolverCache.delete(zipFilePath)
  }

  private static _resolverCache = new Map<string, ZipMapFileResolver>()

  private _loadedZip?: JSZip

  constructor(
    private _zipFilePath: string,
    private _mapFileFromPath: MapFileFromPath =
      ZipMapFileResolver.adjacentMapFileFromPath) {
    super()
  }

  async resolveSourcemap(url: string): Promise<ResolveSourcemapResult> {
    if (this._loadedZip == null) await this._loadZip()
    assert(this._loadedZip != null, 'Did not load zip correctly')
    const matchingFile = this._matchUrl(url)
    if (matchingFile == null) return ResolveSourcemapResult.empty
    const json = await matchingFile.async('text')
    return new ResolveSourcemapResult(Sourcemap.fromRawSourcemapJSON(json))
  }

  // @protected for testing only
  protected async _loadZip(): Promise<void> {
    const src = await readFile(this._zipFilePath)
    this._loadedZip = await loadAsync(src)
  }

  // @protected for testing only
  protected _matchUrl(url: string): JSZipObject | null {
    const matches = this._loadedZip!.filter(
      (_relDir, file) => {
        if (file.dir) return false
        return url.includes(file.name)
      }
    )
    if (matches == null || matches.length === 0) return null
    const generatedFile = matches[0]
    const mapFilePath = this._mapFileFromPath(generatedFile.name)
    return this._loadedZip!.file(mapFilePath)
  }
}