import proxyquire from 'proxyquire'
import test from 'tape'
import { ResolveSourcemapResult } from '../src/SourcemapResolver'

class FsUtilsStub {
  constructor(
    private _accessibleEntries: Set<string> = new Set(),
    private _sourcemaps: Map<string, any> = new Map()
  ) {}

  canRead = (url: string) =>
    Promise.resolve(this._accessibleEntries.has(url))

  isAccessibleDirectory = (dir: string) =>
    Promise.resolve(this._accessibleEntries.has(dir))

  readFile = (url: string) => Promise.resolve(
      this._sourcemaps.has(url) ? this._sourcemaps.get(url) : null
    )
}

class Sourcemap {
  fromRawSourcemapJSON = (json: string) => ( {sourcemap: json} )
}

const ORIG_FILE1 = '/project/file1.ts'
const FILE1 = '/project/file1.js'
const MAP_FILE1 = '/project/file1.js.map'
const FILE1_MAP = JSON.stringify({ file: ORIG_FILE1 })

const FILE2 = '/project/file2.js'

const accessibleFiles = new Set([FILE1, MAP_FILE1])
const sourcemaps = new Map([[MAP_FILE1, FILE1_MAP]])

const {AdjacentMapFileResolver} = proxyquire(
  '../src/AdjacentMapFileResolver', {
    './fs-utils': new FsUtilsStub(accessibleFiles, sourcemaps),
    './Sourcemap': new Sourcemap()
  })

test('adjacent map file: resolve valid url verifying project access',
  async t => {
  try {
    const resolver = new AdjacentMapFileResolver(true)
    const sm: ResolveSourcemapResult = await resolver.resolveSourcemap(FILE1)

    t.ok(!sm.hasError, 'has no error')
    t.ok(!sm.isEmpty, 'is not empty')
    t.ok(sm.sourcemap != null, 'includes sourcemap')
    t.equal(
      sm.sourcemap!.sourcemapData.file,
      ORIG_FILE1,
      'referencing original file'
    )

  } catch (err) {
    t.fail('should not fail ' + err.message)
  } finally {
    t.end()
  }
})

test('adjacent map file: resolve inaccessible url verifying project access',
  async t => {
    try {
      const resolver = new AdjacentMapFileResolver(true)
      await resolver.resolveSourcemap(FILE2)
      t.fail('should throw')
    } catch (err) {
      t.ok(err, 'throws error')
      t.ok(
        (/cannot access project/i).test(err.message),
        'indicating project is inaccessible')
    } finally {
      t.end()
    }
  })

test('adjacent map file: resolve inaccessible url not verifying project access',
  async t => {
    try {
      const resolver = new AdjacentMapFileResolver(false)
      const sm: ResolveSourcemapResult = await resolver.resolveSourcemap(FILE2)
      t.ok(sm.isEmpty, 'returned sourcemap is empty')
    } catch (err) {
      t.fail('should not fail ' + err.message)
    } finally {
      t.end()
    }
  })
