import assert from 'assert'
import proxyquire from 'proxyquire'
import test from 'tape'
import {
  RemoteFileResolver as RemoteFileResolverType
} from '../src/RemoteFileResolver'
import { ResolveSourcemapFailureReason } from '../src/SourcemapResolver'

const ENDPOINT = 'http://some/endpoint'

class Fetch {
  request: RequestInit | null = null
  response: Response | null = null

  constructor(private _endpoint: string) { }

  fetch = (endpoint: string, request: RequestInit) => {
    assert.equal(endpoint, this._endpoint, 'fetches correct endpoint')
    this.request = request
    return Promise.resolve(this.response)
  }
}

class Sourcemap {
  fromRawSourcemapJSON = (json: string) => ({ sourcemap: json })
}

function resolveJson(body: object): Promise<object> {
  return Promise.resolve(body)
}

const fetch = new Fetch(ENDPOINT)
const { RemoteFileResolver } = proxyquire('../src/RemoteFileResolver', {
  'cross-fetch': fetch.fetch,
  './Sourcemap': new Sourcemap()
})
const resolver = new RemoteFileResolver(ENDPOINT) as RemoteFileResolverType

const fileUrl = '/some/file/url.js'
const sourcemap = { file: '/some/file/url.ts' }

test('remote resolver: resolve URL and sourcemap found', async t => {
  fetch.response = {
    status: 200, ok: true, json: () => resolveJson({
      error: null,
      result: sourcemap
    })
  } as Response

  try {
    const result = await resolver.resolveSourcemap(fileUrl)

    t.equal(
      fetch.request!.body,
      '{"fileUrl":"/some/file/url.js"}',
      'fetches file info'
    )
    t.ok(!result.hasError, 'no error')
    t.ok(!result.isEmpty, 'is not empty')
    t.ok(result.failureReason == null, 'no failure')
    t.equal(
      result.sourcemap!.sourcemapData.file,
      sourcemap.file,
      'with orig file')

  } catch (err) {
    t.ifError(err)
  }

  t.end()
})

test('remote resolver: resolve URL but no sourcemap found', async t => {
  fetch.response = {
    status: 200, ok: true, json: () => resolveJson({
      error: null,
      result: null
    })
  } as Response

  try {
    const result = await resolver.resolveSourcemap(fileUrl)

    t.equal(
      fetch.request!.body,
      '{"fileUrl":"/some/file/url.js"}',
      'fetches file info'
    )
    t.ok(!result.hasError, 'no error')
    t.ok(result.isEmpty, 'is empty')
  } catch (err) {
    t.ifError(err)
  }

  t.end()
})

test('remote resolver: resolve URL but not ok', async t => {
  fetch.response = {
    status: 200, ok: false, json: () => resolveJson({
      error: null,
      result: null
    })
  } as Response

  try {
    const result = await resolver.resolveSourcemap(fileUrl)
    t.ok(!result.hasError, 'no error')
    t.ok(result.isEmpty, 'is empty')
  } catch (err) {
    t.ifError(err)
  }

  t.end()
})

test('remote resolver: resolve URL but errors', async t => {
  const error = 'Something went wrong'
  fetch.response = {
    status: 200, ok: true, json: () => resolveJson({
      error,
      result: null
    })
  } as Response

  try {
    const result = await resolver.resolveSourcemap(fileUrl)
    t.ok(result.hasError, 'no error')
    t.ok(result.isEmpty, 'is empty')
    t.equal(result.failureReason, ResolveSourcemapFailureReason.unknown)
    t.ok(result.errorMessage!.includes(error),  'includes error message')
  } catch (err) {
    t.equal(err, error, 'throws error')
  }

  t.end()
})

test('remote resolver: resolve URL for non-user file', async t => {
  fetch.response = {
    status: 200, ok: false, json: () => resolveJson({
      error: null,
      result: null
    })
  } as Response

  try {
    const result = await resolver.resolveSourcemap('./hello.js')
    t.ok(!result.hasError, 'no error')
    t.ok(result.isEmpty, 'is empty')
  } catch (err) {
    t.ifError(err)
  }

  t.end()
})

test('remote resolver: resolve URL but 404, project accessible', async t => {
  fetch.response = {
    status: 404, ok: true, json: () => resolveJson({
      projectInaccessible: false
    })
  } as Response

  try {
    const result = await resolver.resolveSourcemap(fileUrl)
    t.ok(!result.hasError, 'no error')
    t.ok(result.isEmpty, 'is empty')
  } catch (err) {
    t.ifError(err)
  }

  t.end()
})

test('remote resolver: resolve URL but 404, project inaccessible', async t => {
  fetch.response = {
    status: 404, ok: true, json: () => resolveJson({
      projectInaccessible: true
    })
  } as Response

  try {
    const result = await resolver.resolveSourcemap(fileUrl)
    t.ok(result.hasError, 'error')
    t.ok(result.isEmpty, 'is empty')
    t.ok((/cannot access the project/i).test(result.errorMessage!), 'indicates project is inaccessible')
    t.equal(result.failureReason,
      ResolveSourcemapFailureReason.projectInaccessible,
      'reason: project inaccessible'
    )
  } catch (err) {
    t.ifError(err)
  }

  t.end()
})

test('remote resolver: resolve URL but not ok', async t => {
  fetch.response = {
    status: 200, ok: false, json: () => resolveJson({})
  } as Response

  try {
    const result = await resolver.resolveSourcemap(fileUrl)
    t.ok(!result.hasError, 'no error')
    t.ok(result.isEmpty, 'is empty')
  } catch (err) {
    t.ifError(err)
  }

  t.end()
})

test('remote resolver: resolve URL with zip file included', async t => {
  const zipResolver = new RemoteFileResolver(ENDPOINT, '/path/to/sourcemap.zip') as RemoteFileResolverType

  fetch.response = {
    status: 200, ok: true, json: () => resolveJson({
      error: null,
      result: sourcemap
    })
  } as Response

  try {
    const result = await zipResolver.resolveSourcemap(fileUrl)

    t.equal(
      fetch.request!.body,
      '{"fileUrl":"/some/file/url.js","zipMapFile":"/path/to/sourcemap.zip"}',
      'fetches file info including zip file location'
    )
    t.ok(!result.hasError, 'no error')
    t.ok(!result.isEmpty, 'is not empty')
    t.ok(result.failureReason == null, 'no failure')
    t.equal(
      result.sourcemap!.sourcemapData.file,
      sourcemap.file,
      'with orig file')

  } catch (err) {
    t.ifError(err)
  }

  t.end()
})

