import assert from 'assert'
import path from 'path'
import { RawSourceMap } from 'source-map'
import spok from 'spok'
import test from 'tape'

import { CpuProfileNode } from '../src/CpuProfile'
import { SourceInfoMapper } from '../src/SourceInfoMapper'
import { Sourcemap } from '../src/Sourcemap'
import {
  ResolveSourcemapFailureReason,
  ResolveSourcemapResult,
  SourcemapResolver
} from '../src/SourcemapResolver'

function cpuProfileNode(url: string, children: CpuProfileNode[] = [])
  : CpuProfileNode {
  return {
    functionName: '<anonymous>',
    url,
    callUUID: 0,
    bailoutReason: 'none',
    id: 1,
    scriptId: 2,
    lineNumber: 1,
    hitCount: 0,
    children
  }
}

function cpuProfileNodesFor(urls: string[]) {
  assert(urls != null && urls.length > 0, 'need at least one URL my friend')
  const nodesMap: Map<string, CpuProfileNode> = new Map()

  const childNodes = urls.slice(1).map(url => cpuProfileNode(url))
  const rootNode = cpuProfileNode(urls[0], childNodes)

  nodesMap.set(urls[0], rootNode)
  for (const node of childNodes) {
    nodesMap.set(node.url, node)
  }

  return {
    nodesMap,
    nodes: [rootNode]
  }
}

function sourceMap(url: string): Sourcemap {
  const rawSourceMap: RawSourceMap = {
    version: 3,
    sources: [`mapped:${path.basename(url)}`],
    names: ['var names'],
    mappings: 'AAAA',
    file: url
  }
  return Sourcemap.fromRawSourcemap(rawSourceMap)
}

function successResolveSourcemapResult(url: string) {
  return new ResolveSourcemapResult(sourceMap(url))
}

class SourcemapResolverMock extends SourcemapResolver {
  resolvedURLs: string[] = []

  constructor(
    private resolutionResults: Map<string, ResolveSourcemapResult> = new Map(),
    private resolutionErrors: Set<string> = new Set()
  ) { super() }

  async resolveSourcemap(url: string): Promise<ResolveSourcemapResult> {
    this.resolvedURLs.push(url)

    if (this.resolutionErrors.has(url)) {
      return Promise.resolve(
        ResolveSourcemapResult.failureResult(
          ResolveSourcemapFailureReason.unknown,
          'Resolving ' + url
        )
      )
    }

    const result = this.resolutionResults.has(url)
      ? this.resolutionResults.get(url)
      : ResolveSourcemapResult.empty

    return Promise.resolve(result!)
  }
}

const URL1 = '/some/url/file1.js'
const URL2 = '/some/url/file2.js'
const URL3 = '/some/url/file3.js'
const URL4 = '/some/url/file4.js'
const URL5 = '/some/url/file5.js'

async function runMap(
  urls: string[],
  resolutionResults: Map<string, ResolveSourcemapResult> = new Map(),
  resolutionErrors: Set<string> = new Set()
) {
  const resolver = new SourcemapResolverMock(
    resolutionResults, resolutionErrors
  )
  const sut = SourceInfoMapper.forResolver(resolver)
  const { nodes, nodesMap } = cpuProfileNodesFor(urls)
  const errors = await sut.map(nodes)

  return { errors, resolver, nodesMap }
}

function urlArray(nodesMap: Map<string, CpuProfileNode>):
  Array<[string, string]> {
  const map: Map<string, string> = new Map()
  for (const [key, node] of nodesMap) {
    map.set(key, node.url)
  }

  return Array.from(map)
}

test('sourceinfomapper: resolve one unknown URL', async t => {
  try {
    const urls = [URL1]
    const { errors, resolver } = await runMap(urls)
    t.deepEqual(resolver.resolvedURLs, urls, 'resolves all contained URLs')
    t.equal(errors.length, 0, 'returns no errors')
    t.end()
  } catch (err) {
    t.fail('should not fail ' + err.message)
  }
})

test('sourceinfomapper: resolve 2 unknown URLs and 3 known ones', async t => {
  try {
    const urls = [URL1, URL2, URL3, URL4, URL5]
    const resolutionResults = new Map([
      [URL3, successResolveSourcemapResult(URL3)],
      [URL4, successResolveSourcemapResult(URL4)],
      [URL5, successResolveSourcemapResult(URL5)]
    ])

    const { errors, resolver, nodesMap } = await runMap(urls, resolutionResults)
    t.deepEqual(resolver.resolvedURLs, urls, 'resolves all contained URLs')
    t.equal(errors.length, 0, 'returns no errors')
    spok(t, urlArray(nodesMap), [
      ['/some/url/file1.js', '/some/url/file1.js'],
      ['/some/url/file2.js', '/some/url/file2.js'],
      ['/some/url/file3.js', '/some/url/mapped:file3.js'],
      ['/some/url/file4.js', '/some/url/mapped:file4.js'],
      ['/some/url/file5.js', '/some/url/mapped:file5.js']
    ]
    )
    t.end()
  } catch (err) {
    t.fail('should not fail ' + err.message)
  }
})

test('sourceinfomapper: resolve 2 unknown URLs and 1 known one and 2 errors',
  async t => {
    try {
      const urls = [URL1, URL2, URL3, URL4, URL5]
      const resolutionResults = new Map([
        [URL3, successResolveSourcemapResult(URL3)]
      ])
      const resolutionErrors = new Set([URL4, URL5])

      const { errors, resolver, nodesMap } =
        await runMap(urls, resolutionResults, resolutionErrors)
      t.deepEqual(resolver.resolvedURLs, urls, 'resolves all contained URLs')
      t.equal(errors.length, 2, 'returns two errors')
      spok(t, urlArray(nodesMap), [
        ['/some/url/file1.js', '/some/url/file1.js'],
        ['/some/url/file2.js', '/some/url/file2.js'],
        ['/some/url/file3.js', '/some/url/mapped:file3.js']
      ]
      )
      spok(t, errors.map(x => x.errorMessage), [
        'Resolving /some/url/file4.js',
        'Resolving /some/url/file5.js'
      ]
      )
      t.end()
    } catch (err) {
      t.fail('should not fail ' + err.message)
    }
  })
