import JSZip, { JSZipObject } from 'jszip'
import test from 'tape'
import {
  ZipMapFileResolver
} from '../src/ZipMapFileResolver'

import {
  RawSourceMap
} from 'source-map'

const URL = '/some/file.js'
const sourcemap: RawSourceMap = {
  version: 3,
  sources: ['some/file.tx'],
  names: [],
  mappings: 'AAAA',
  file: URL
}

// Overriding part of the class under test b/c stubbing out
// the JSZip module is too complex.
class ZipMapFileResolverTestable extends ZipMapFileResolver {
  matchedURL: string | null = null
  loadedZip: boolean = false

  constructor(
    private _tobeLoadedZip: JSZip | null,
    private _tobeMatchedFile: JSZipObject | null
  ) {
    super('')
  }
  protected _loadZip(): Promise<void> {
    // @ts-ignore
    this._loadedZip = this._tobeLoadedZip
    this.loadedZip = true
    return Promise.resolve()
  }

  protected _matchUrl(url: string): JSZipObject | null {
    this.matchedURL = url
    return this._tobeMatchedFile
  }
}

const LOADED_ZIP = {} as JSZip
// @ts-ignore
const MATCHING_FILE = {
  async() {
    return Promise.resolve(JSON.stringify(sourcemap))
  }
} as JSZipObject

test('zip map file resolver: first load, loading zip fails', async t => {
  const resolver = new ZipMapFileResolverTestable(null, null)
  try {
    await resolver.resolveSourcemap(URL)
    t.fail('should throw')
  } catch (err) {
    t.ok(resolver.loadedZip, 'loads zip')
    t.ok(err, 'throws assertion error')
    t.ok(
      (/did not load zip correctly/i).test(err.message),
      'indicating problem')
  } finally {
    t.end()
  }
})

test('zip map file resolver: first load, no matching file', async t => {
  try {
    const resolver = new ZipMapFileResolverTestable(LOADED_ZIP, null)
    const result = await resolver.resolveSourcemap(URL)

    t.ok(resolver.loadedZip, 'loads zip')
    t.equal(resolver.matchedURL, URL, 'matches given URL')
    t.ok(result.isEmpty, 'is empty')
    t.ok(!result.hasError, 'no error')
  } catch (err) {
    t.fail('should not fail ' + err.message)
  } finally {
    t.end()
  }
})

test('zip map file resolver: first load, matching file', async t => {
  try {
    const resolver = new ZipMapFileResolverTestable(LOADED_ZIP, MATCHING_FILE)
    const result = await resolver.resolveSourcemap(URL)

    t.ok(resolver.loadedZip, 'loads zip')
    t.equal(resolver.matchedURL, URL, 'matches given URL')
    t.ok(!result.isEmpty, 'not empty')
    t.ok(!result.hasError, 'no error')
    t.equal(
      result.sourcemap!.sourcemapData.file,
      sourcemap.file,
      'with orig file')
  } catch (err) {
    t.fail('should not fail ' + err.message)
  } finally {
    t.end()
  }
})

test('zip map file resolver: second load, matching file', async t => {
  try {
    const resolver = new ZipMapFileResolverTestable(LOADED_ZIP, MATCHING_FILE)
    // @ts-ignore
    resolver._loadedZip = LOADED_ZIP
    const result = await resolver.resolveSourcemap(URL)

    t.ok(!resolver.loadedZip, 'does not load zip again')
    t.equal(resolver.matchedURL, URL, 'matches given URL')
    t.ok(!result.isEmpty, 'not empty')
    t.ok(!result.hasError, 'no error')
    t.equal(
      result.sourcemap!.sourcemapData.file,
      sourcemap.file,
      'with orig file')
  } catch (err) {
    t.fail('should not fail ' + err.message)
  } finally {
    t.end()
  }
})
