#!/bin/bash

ARCH=$(uname)
NVM=~/.nvm/nvm.sh
DOCKER=$(which docker)
SNAPPDIFF_VERSION=${SNAPDIFF_VERSION:=$1}

# Check if snapdiff version exists
if [[ -z $SNAPDIFF_VERSION ]]; then
	echo "Please include a SNAPPDIFF_VERSION either as an environment"
	echo "variable or the first argument to the script."
	exit 1
fi

if [[ -d ./snapdiff ]]; then
	rm -rf ./snapdiff
fi

git clone --branch $SNAPDIFF_VERSION --depth 1 git@github.com:nodesource/snapdiff.git
mv $(npm pack ./snapdiff | tail -1) ./snapdiff.tgz

# Check if OSX
if [[ $ARCH != 'Darwin' ]]; then
  echo "This script can only be run on Darwin. Please try again"
  exit 1
fi

# Check if NVM installed
if [[ ! -f $NVM ]]; then
  echo "Node Version Manager is needed to run this script. Please install it with:"
  echo "curl -o- https://raw.githubusercontent.com/creationix/nvm/v0.33.6/install.sh | bash"
  exit 1
fi
. ~/.nvm/nvm.sh
. ~/.bashrc

# Check if Docker installed
if [[ -z $DOCKER ]]; then
  echo "Docker is needed to run this script. Get it here:"
  echo "https://www.docker.com/community-edition#/download"
  exit 1
fi

echo ""
echo "---"
echo "Starting prebuild updates..."
echo "---"
echo ""

echo "Logging into AWS ECR..."
`aws ecr get-login --no-include-email`

SCRIPTDIR=$(
  cd $(dirname "$0")
  pwd
)

if [[ -d $SCRIPTDIR/output ]]; then
  rm -rf $SCRIPTDIR/output
fi

mkdir -p $SCRIPTDIR/output
cd $SCRIPTDIR/output
rsync -rvq --exclude=output --exclude=lib ../../* ./
mkdir -p lib

echo ""
echo "---"
echo "Starting Darwin..."
echo "---"
echo ""

nvm install 8
node install.js ./snapdiff

nvm install 10
node install.js ./snapdiff

nvm install 12
node install.js ./snapdiff

find lib/ -name .npmignore -exec rm -rf {} \;
rm -rf ../../lib/darwin*
mv lib/darwin* ../../lib/

echo ""
echo "---"
echo "Darwin Complete."
echo "---"
echo ""
echo "---"
echo "Starting Linux x64..."
echo "---"
echo ""

docker run -v $(pwd):/root/snapdiff-build --entrypoint /root/snapdiff-build/tools/prebuild-scripts/update-x64.sh 800406105498.dkr.ecr.us-west-2.amazonaws.com/codebuild:build-binary-linux

docker run -v $(pwd):/root/snapdiff-build --entrypoint /root/snapdiff-build/tools/prebuild-scripts/update-x64-erbium.sh 800406105498.dkr.ecr.us-west-2.amazonaws.com/codebuild:build-binary-linux-centos7

find lib/ -name .npmignore -exec rm -rf {} \;
rm -rf ../../lib/linux-x64*
mv lib/linux-x64* ../../lib/

echo ""
echo "---"
echo "Linux x64 Complete."
echo "---"
echo ""
echo "---"
echo "Starting Linux ARM..."
echo "---"
echo ""

docker run -v $(pwd):/root/snapdiff-build --entrypoint /root/snapdiff-build/tools/prebuild-scripts/update-arm.sh 800406105498.dkr.ecr.us-west-2.amazonaws.com/codebuild:build-binary-arm

find lib/ -name .npmignore -exec rm -rf {} \;
rm -rf ../../lib/linux-arm*
mv lib/linux-arm* ../../lib/

echo ""
echo "---"
echo "Linux ARM Complete."
echo "---"
echo ""
echo "---"
echo "Starting Alpine Linux [NodeJS Carbon]..."
echo "---"
echo ""

docker run -v $(pwd):/root/snapdiff-build --entrypoint /root/snapdiff-build/tools/prebuild-scripts/update-alpine.sh 800406105498.dkr.ecr.us-west-2.amazonaws.com/codebuild:build-binary-alpine-carbon

find lib/ -name .npmignore -exec rm -rf {} \;
rm -rf ../../lib/alpine-x64-57
mv lib/alpine* ../../lib/

echo ""
echo "---"
echo "Alpine Linux [NodeJS Carbon] Complete."
echo "---"
echo ""
echo "---"
echo "Starting Alpine Linux [NodeJS Dubnium]..."
echo "---"
echo ""

rm -rf $SCRIPTDIR/output

docker run -v $(pwd):/root/snapdiff-build --entrypoint /root/snapdiff-build/tools/prebuild-scripts/update-alpine.sh 800406105498.dkr.ecr.us-west-2.amazonaws.com/codebuild:build-binary-alpine-dubnium

find lib/ -name .npmignore -exec rm -rf {} \;
rm -rf ../../lib/alpine-x64-64
mv lib/alpine* ../../lib/

echo ""
echo "---"
echo "Alpine Linux [NodeJS Dubnium] Complete."
echo "---"
echo ""
echo "---"
echo "Starting Alpine Linux [NodeJS Erbium]..."
echo "---"
echo ""

rm -rf $SCRIPTDIR/output

docker run -v $(pwd):/root/snapdiff-build --entrypoint /root/snapdiff-build/tools/prebuild-scripts/update-alpine.sh node:erbium-alpine

find lib/ -name .npmignore -exec rm -rf {} \;
rm -rf ../../lib/alpine-x64-64
mv lib/alpine* ../../lib/

echo ""
echo "---"
echo "Alpine Linux [NodeJS Erbium] Complete."
echo "---"
echo ""

rm -rf $SCRIPTDIR/output
echo ""
echo "---"
echo "All prebuilt native modules have been updated."
echo "---"
echo ""
exit 0

